__all__ = ["DistributionChart"]

import typing

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FFMpegWriter, FuncAnimation, ImageMagickWriter
from PySide6 import QtCore, QtGui, QtWidgets
from scipy.stats import pearsonr

from ..ssu import SSUResult, SSUViewModel
from ..statistic import convert_φ_to_μm, get_mode
from .BaseChart import BaseChart
from .config_matplotlib import normal_color

class DistributionChart(BaseChart):
    def __init__(self, parent=None, figsize=(4, 4)):
        super().__init__(parent=parent, figsize=figsize)
        self.axes = self.figure.subplots() # type: plt.Axes

        self.scale_menu = QtWidgets.QMenu(self.tr("Scale")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name) # type: QtGui.QAction
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)
        self.show_mode_lines_action = QtGui.QAction(self.tr("Show Mode Lines")) # type: QtGui.QAction
        self.show_mode_lines_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.show_mode_lines_action)
        self.show_mode_lines_action.setCheckable(True)
        self.show_mode_lines_action.setChecked(False)
        self.show_legend_action = QtGui.QAction(self.tr("Show Legend")) # type: QtGui.QAction
        self.show_legend_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.show_legend_action)
        self.show_legend_action.setCheckable(True)
        self.show_legend_action.setChecked(False)
        self.animated_action = QtGui.QAction(self.tr("Animated")) # type: QtGui.QAction
        self.animated_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.animated_action)
        self.animated_action.setCheckable(True)
        self.animated_action.setChecked(False)
        self.interval_menu = QtWidgets.QMenu(self.tr("Animation Interval")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.interval_menu)
        self.interval_group = QtGui.QActionGroup(self.interval_menu)
        self.interval_group.setExclusive(True)
        self.interval_actions = [] # type: list[QtGui.QAction]
        for interval, name in self.supported_intervals:
            interval_action = self.interval_group.addAction(name)
            interval_action.setCheckable(True)
            interval_action.triggered.connect(self.update_chart)
            self.interval_menu.addAction(interval_action)
            self.interval_actions.append(interval_action)
        self.interval_actions[3].setChecked(True)
        self.repeat_animation_action = QtGui.QAction(self.tr("Repeat Animation")) # type: QtGui.QAction
        self.repeat_animation_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.repeat_animation_action)
        self.repeat_animation_action.setCheckable(True)
        self.repeat_animation_action.setChecked(False)
        self.save_animation_action = QtGui.QAction(self.tr("Save Animation")) # type: QtGui.QAction
        self.menu.addAction(self.save_animation_action)
        self.save_animation_action.triggered.connect(self.save_animation)

        self.animation = None
        self.last_model = None # type: SSUViewModel
        self.last_result = None # type: SSUResult
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

    @property
    def supported_scales(self) -> typing.List[typing.Tuple[str, str]]:
        scales = [("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear"))]
        return scales

    @property
    def supported_intervals(self) -> typing.List[typing.Tuple[int, str]]:
        intervals = [(5, self.tr("5 Milliseconds")),
                     (10, self.tr("10 Milliseconds")),
                     (20, self.tr("20 Milliseconds")),
                     (30, self.tr("30 Milliseconds")),
                     (60, self.tr("60 Milliseconds"))]
        return intervals

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def show_mode_lines(self) -> bool:
        return self.show_mode_lines_action.isChecked()

    @property
    def show_legend(self) -> bool:
        return self.show_legend_action.isChecked()

    @property
    def animated(self) -> bool:
        return self.animated_action.isChecked()

    @property
    def animation_interval(self) -> int:
        for i, interval_action in enumerate(self.interval_actions):
            if interval_action.isChecked():
                interval, name = self.supported_intervals[i]
                return interval

    @property
    def repeat_animation(self) -> bool:
        return self.repeat_animation_action.isChecked()

    @property
    def transfer(self) -> typing.Callable:
        if self.scale == "log-linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)
        elif self.scale == "log":
            return lambda classes_φ: np.log(convert_φ_to_μm(classes_φ))
        elif self.scale == "phi":
            return lambda classes_φ: classes_φ
        elif self.scale == "linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size [μm]"
        elif self.scale == "log":
            return "Ln(grain size in μm)"
        elif self.scale == "phi":
            return "Grain size [φ]"
        elif self.scale == "linear":
            return "Grain size [μm]"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    @property
    def repeat(self) -> bool:
        return self.repeat_check_box.isChecked()

    def show_menu(self, pos: QtCore.QPoint):
        self.save_figure_action.setEnabled(self.last_model is not None)
        self.save_animation_action.setEnabled(self.last_result is not None)
        self.menu.popup(QtGui.QCursor.pos())

    def show_model(self, model: SSUViewModel, quick=False):
        modes_φ = [model.classes_φ[np.unravel_index(np.argmax(distribution), distribution.shape)] for distribution in model.distributions]
        if self.animation is not None:
            self.animation._stop()
            self.animation = None
        if not quick:
            self.last_result = None
            self.last_model = model
            self.axes.clear()
            x = self.transfer(model.classes_φ)
            self.axes.set_title(model.title)
            self.axes.set_xlabel(self.xlabel)
            self.axes.set_ylabel(self.ylabel)
            self.target_line = self.axes.plot(x, model.target, c="#ffffff00", marker=".", ms=8, mfc=normal_color(), mew=0.0)[0]
            self.axes.set_xlim(x[0], x[-1])
            self.axes.set_ylim(0.0, round(np.max(model.target)*1.2, 2))
            self.mixed_line = self.axes.plot(x, model.mixed, c=normal_color())[0]
            self.component_lines = []
            for i, (distribution, proportion) in enumerate(zip(model.distributions, model.proportions)):
                component = self.axes.plot(x, distribution*proportion, c=plt.get_cmap()(i))[0]
                self.component_lines.append(component)
            self.figure.tight_layout()
        else:
            self.target_line.set_ydata(model.target)
            self.mixed_line.set_ydata(model.mixed)
            for component_line, distribution, proportion in zip(self.component_lines, model.distributions, model.proportions):
                component_line.set_ydata(distribution*proportion)
        if self.xlog:
            self.axes.set_xscale("log")
        if self.show_mode_lines:
            if hasattr(self, "vlines"):
                self.vlines.remove()
            modes = [self.transfer(mode_φ) for mode_φ in modes_φ]
            colors = [plt.get_cmap()(i) for i in range(model.n_components)]
            self.vlines = self.axes.vlines(modes, 0.0, 1.0, colors=colors)
        if self.show_legend:
            r, p = pearsonr(model.target, model.mixed)
            r2 = r**2
            handles = [self.target_line, self.mixed_line]
            handles.extend(self.component_lines)
            labels = ["Target", f"Mixed ($R^2$={r2:.2f})"]
            for i, (mode_φ, proportion) in enumerate(zip(modes_φ, model.proportions)):
                mode_μm = convert_φ_to_μm(mode_φ)
                label = f"{model.component_prefix}{i+1} ({mode_μm:.2f} μm, {proportion:.2%})"
                labels.append(label)
            self.legend = self.axes.legend(
                handles=handles, labels=labels,
                loc="upper left", prop={"size": 8})
        self.canvas.draw()

    def show_animation(self, result: SSUResult):
        if self.animation is not None:
            self.animation._stop()
            self.animation = None
        self.last_model = None
        self.last_result = result

        models = iter(result.view_models)
        first = next(models)
        x = self.transfer(first.classes_φ)
        self.axes.cla()
        if self.xlog:
            self.axes.set_xscale("log")
        self.axes.set_title(first.title)
        self.axes.set_xlabel(self.xlabel)
        self.axes.set_ylabel(self.ylabel)
        self.target_line = self.axes.plot(x, first.target, c="#ffffff00", marker=".", ms=8, mfc=normal_color(), mew=0.0)[0]
        self.axes.set_xlim(x[0], x[-1])
        self.axes.set_ylim(0.0, round(np.max(first.target)*1.2, 2))
        self.figure.tight_layout()
        # self.canvas.draw()
        def common(model):
            modes_φ = [model.classes_φ[np.unravel_index(np.argmax(distribution), distribution.shape)] for distribution in model.distributions]
            if self.show_mode_lines:
                if hasattr(self, "vlines"):
                    self.vlines.remove()
                modes = [self.transfer(mode_φ) for mode_φ in modes_φ]
                colors = [plt.get_cmap()(i) for i in range(model.n_components)]
                self.vlines = self.axes.vlines(modes, 0.0, 1.0, colors=colors)
            if self.show_legend:
                r, p = pearsonr(model.target, model.mixed)
                r2 = r**2
                handles = [self.target_line, self.mixed_line]
                handles.extend(self.component_lines)
                labels = ["Target", f"Mixed ($R^2$={r2:.2f})"]
                for i, (mode_φ, proportion) in enumerate(zip(modes_φ, model.proportions)):
                    mode_μm = convert_φ_to_μm(mode_φ)
                    label = f"{model.component_prefix}{i+1} ({mode_μm:.2f} μm, {proportion:.2%})"
                    labels.append(label)
                self.legend = self.axes.legend(
                    handles=handles, labels=labels,
                    loc="upper left", prop={"size": 8})
        def init():
            model = first
            self.mixed_line = self.axes.plot(x, model.mixed, c=normal_color())[0]
            self.component_lines = [self.axes.plot(x, distribution*proportion, c=plt.get_cmap()(i))[0] for i, (distribution, proportion) in enumerate(zip(model.distributions, model.proportions))]
            common(model)
            check_artists = [self.mixed_line]
            check_artists.extend(self.component_lines)
            if self.show_mode_lines:
                check_artists.append(self.vlines)
            if self.show_legend:
                check_artists.append(self.legend)
            return check_artists
        def animate(current: SSUViewModel):
            model = current
            self.mixed_line.set_ydata(current.mixed)
            for line, distribution, proportion in zip(self.component_lines, model.distributions, model.proportions):
                line.set_ydata(distribution*proportion)
            common(model)
            check_artists = [self.mixed_line]
            check_artists.extend(self.component_lines)
            if self.show_mode_lines:
                check_artists.append(self.vlines)
            if self.show_legend:
                check_artists.append(self.legend)
            return check_artists
        self.animation = FuncAnimation(
            self.figure, animate, frames=models, init_func=init,
            interval=self.animation_interval, blit=True,
            repeat=self.repeat_animation, repeat_delay=3.0, save_count=result.n_iterations)

    def show_result(self, result: SSUResult):
        if self.animated:
            self.show_animation(result)
        else:
            self.show_model(result.view_model)
        self.last_model = None
        self.last_result = result

    def save_animation(self):
        if self.last_result is not None:
            filename, format_str = self.file_dialog.getSaveFileName(
                self, self.tr("Choose a filename to save the animation of this SSU result"),
                None, "MPEG-4 Video File (*.mp4);;Graphics Interchange Format (*.gif)")
            if filename is None or filename == "":
                return
            progress = QtWidgets.QProgressDialog(self)
            progress.setWindowTitle("QGrain")
            progress.setRange(0, 100)
            progress.setLabelText(self.tr("Saving Animation [{0} Frames]").format(self.last_result.n_iterations))
            canceled = False
            def save_callback(i, n):
                if progress.wasCanceled():
                    nonlocal canceled
                    canceled = True
                    raise StopIteration()
                progress.setValue((i+1)/n*100)
                QtCore.QCoreApplication.processEvents()
            self.show_animation(self.last_result)
            # plt.rcParams["savefig.dpi"] = 120.0
            if "*.gif" in format_str:
                if not ImageMagickWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("ImageMagick is not installed, please download and install it from its offical website (https://imagemagick.org/index.php)."))
                    self.normal_msg.exec_()
                else:
                    self.animation.save(filename, writer="imagemagick", fps=30, progress_callback=save_callback)
            elif "*.mp4" in format_str:
                if not FFMpegWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("FFMpeg is not installed, please download and install it from its offical website (https://ffmpeg.org/)."))
                    self.normal_msg.exec_()
                else:
                    self.animation.save(filename, writer="ffmpeg", fps=30, progress_callback=save_callback)
            # plt.rcParams["savefig.dpi"] = 300.0
            if not canceled:
                progress.setValue(100)

    def update_chart(self):
        self.figure.clear()
        self.axes = self.figure.subplots()
        if self.last_model is not None:
            self.show_model(self.last_model)
        elif self.last_result is not None:
            self.show_result(self.last_result)

    def retranslate(self):
        super().retranslate()
        self.scale_menu.setTitle(self.tr("Scale"))
        for action, (key, name) in zip(self.scale_actions, self.supported_scales):
            action.setText(name)
        self.show_mode_lines_action.setText(self.tr("Show Mode Lines"))
        self.show_legend_action.setText(self.tr("Show Legend"))
        self.animated_action.setText(self.tr("Animated"))
        self.interval_menu.setTitle(self.tr("Animation Interval"))
        for action, (interval, name) in zip(self.interval_actions, self.supported_intervals):
            action.setText(name)
        self.repeat_animation_action.setText(self.tr("Repeat Animation"))
        self.save_animation_action.setText(self.tr("Save Animation"))



__all__ = ["ClusteringAnalyzer"]

import logging

from PySide6 import QtCore, QtWidgets
from scipy.cluster.hierarchy import fcluster, linkage

from ..chart.HierarchicalChart import HierarchicalChart
from ..io import save_clustering
from ..model import GrainSizeDataset


class ClusteringAnalyzer(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.ClusteringAnalyzer")
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.init_ui()
        self.file_dialog = QtWidgets.QFileDialog(self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.__dataset = None
        self.__distribution_matrix = None
        self.__last_result = None

    def init_ui(self):
        self.setWindowTitle(self.tr("Clustering Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.chart = HierarchicalChart()
        self.main_layout.addWidget(self.chart, 0, 0, 1, 8)

        self.supported_methods = [
            "single", "complete",
            "average", "weighted",
            "centroid", "median",
            "ward"]
        self.supported_distances = [
            "braycurtis", "canberra",
            "chebyshev", "cityblock",
            "correlation", "cosine",
            "dice", "euclidean",
            "hamming", "jaccard",
            "jensenshannon", "kulsinski",
            "mahalanobis", "matching",
            "minkowski", "rogerstanimoto",
            "russellrao", "seuclidean",
            "sokalmichener", "sokalsneath",
            "sqeuclidean", "yule"]

        self.linkage_method_label = QtWidgets.QLabel(self.tr("Linkage Method"))
        self.linkage_method_label.setToolTip(self.tr("The linkage method for calculating the distance between the newly formed cluster and each observation vector."))
        self.linkage_method_combo_box = QtWidgets.QComboBox()
        self.linkage_method_combo_box.addItems(self.supported_methods)
        self.linkage_method_combo_box.setCurrentText("ward")
        self.linkage_method_combo_box.currentTextChanged.connect(lambda linkage: self.perform())
        self.distance_label = QtWidgets.QLabel(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The distance metric."))
        self.distance_combo_box = QtWidgets.QComboBox()
        self.distance_combo_box.addItems(self.supported_distances)
        self.distance_combo_box.setCurrentText("euclidean")
        self.distance_combo_box.currentTextChanged.connect(lambda distance: self.perform())
        self.main_layout.addWidget(self.linkage_method_label, 1, 0)
        self.main_layout.addWidget(self.linkage_method_combo_box, 1, 1)
        self.main_layout.addWidget(self.distance_label, 1, 2)
        self.main_layout.addWidget(self.distance_combo_box, 1, 3)

        self.p_label = QtWidgets.QLabel(self.tr("p"))
        self.p_label.setToolTip(self.tr("Controls the number of leaves at the bottom level of the figure."))
        self.p_input = QtWidgets.QSpinBox()
        self.p_input.setMinimum(1)
        self.p_input.valueChanged.connect(lambda p: self.perform())
        self.main_layout.addWidget(self.p_label, 1, 4)
        self.main_layout.addWidget(self.p_input, 1, 5)
        self.n_clusers_label = QtWidgets.QLabel(self.tr("Number of Clusters"))
        self.n_clusers_label.setToolTip(self.tr("Controls the number of clusters of this clustering algorithm."))
        self.n_clusers_input = QtWidgets.QSpinBox()
        self.n_clusers_input.setMinimum(2)
        self.main_layout.addWidget(self.n_clusers_label, 1, 6)
        self.main_layout.addWidget(self.n_clusers_input, 1, 7)

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def linkage_method(self) -> str:
        return self.linkage_method_combo_box.currentText()

    @property
    def distance(self) -> str:
        return self.distance_combo_box.currentText()

    @property
    def p(self) -> int:
        return self.p_input.value()

    @property
    def n_clusters(self) -> int:
        return self.n_clusers_input.value()

    def on_dataset_loaded(self, dataset: GrainSizeDataset):
        if dataset is None or not dataset.has_sample:
            return
        self.__dataset = dataset
        self.__distribution_matrix = dataset.distribution_matrix
        self.__last_result = None

        self.p_input.setMaximum(dataset.n_samples)
        self.n_clusers_input.setMaximum(dataset.n_samples-1)
        if dataset.n_samples > 20:
            self.p_input.setValue(10)

        self.perform()

    def perform(self):
        try:
            if self.__last_result is not None:
                self.logger.debug("The clustering algorithm has been performed on this dataset, use previous result to avoid the repetitive computation.")
                linkage_method, distance, linkage_matrix = self.__last_result
                if linkage_method == self.linkage_method and distance == self.distance:
                    self.chart.show_result(linkage_matrix, p=self.p)
                    return
            self.logger.debug(f"Calculate the linkage matrix with the method ({self.linkage_method}) and metric ({self.distance}).")
            linkage_matrix = linkage(self.__distribution_matrix, method=self.linkage_method, metric=self.distance)
            self.__last_result = (self.linkage_method, self.distance, linkage_matrix)
            self.chart.show_result(linkage_matrix, p=self.p)
        except ValueError as e:
            self.logger.error("The linkage method is not compatible with the distance metric.")
            self.show_error(self.tr("The linkage method is not compatible with the distance metric."))
            return

    def save_result(self):
        if self.__last_result is None:
            self.logger.error("The clustering algorithm has not been performed.")
            self.show_error(self.tr("The clustering algorithm has not been performed."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            None, self.tr("Choose a filename to save the clustering result"),
            None, "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        try:
            linkage_method, distance, linkage_matrix = self.__last_result
            flags = fcluster(linkage_matrix, self.n_clusters, criterion="maxclust")
            self.logger.debug(f"Try to save the clustering result. Linkage method: {linkage_method}. Metric: {distance}. Number of clusters: {self.n_clusters}.")
            progress_dialog = QtWidgets.QProgressDialog(
                self.tr("Saving clustering result..."), self.tr("Cancel"),
                0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            save_clustering(self.__dataset, flags, filename, progress_callback=callback, logger=self.logger)
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Clustering Analyzer"))
        self.linkage_method_label.setText(self.tr("Linkage Method"))
        self.linkage_method_label.setToolTip(self.tr("The linkage method for calculating the distance between the newly formed cluster and each observation vector."))
        self.distance_label.setText(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The distance metric."))
        self.p_label.setText(self.tr("p"))
        self.p_label.setToolTip(self.tr("Controls the number of leaves at the bottom level of the figure."))
        self.n_clusers_label.setText(self.tr("Number of Clusters"))
        self.n_clusers_label.setToolTip(self.tr("Controls the number of clusters of this clustering algorithm."))



import typing

import matplotlib.pyplot as plt

from .BaseChart import BaseChart


class DistanceCurveChart(BaseChart):
    def __init__(self, parent=None, figsize=(4, 3)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Distance Series"))
        self.axes = self.figure.subplots()
        self.__last_result = None

    def show_distance_series(self, series: typing.Iterable[float], ylabel: str, title: str = ""):
        self.axes.clear()
        self.axes.plot(series)
        self.axes.set_xlabel("Iteration index")
        self.axes.set_ylabel(ylabel)
        self.axes.set_title(title)
        self.figure.tight_layout()
        self.canvas.draw()
        self.__last_result = series, ylabel, title

    def update_chart(self):
        if self.__last_result is not None:
            self.figure.clear()
            self.axes = self.figure.subplots()
            self.show_distance_series(*self.__last_result)

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("Distance Series"))



__all__ = ["BaseChart"]

import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from PySide6 import QtCore, QtGui, QtWidgets

from .config_matplotlib import setup_matplotlib


class BaseChart(QtWidgets.QWidget):
    def __init__(self, parent=None, figsize=(4, 3)):
        super().__init__(parent=parent)
        self.figure = plt.figure(figsize=figsize)
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.canvas, 0, 0)

        self.menu = QtWidgets.QMenu(self.canvas)
        self.menu.setShortcutAutoRepeat(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_menu)
        self.save_figure_action = self.menu.addAction(self.tr("Save Figure"))
        self.save_figure_action.triggered.connect(lambda: self.toolbar.save_figure())
        self.normal_msg = QtWidgets.QMessageBox(parent=self)

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def show_menu(self, pos: QtCore.QPoint):
        self.menu.popup(QtGui.QCursor.pos())

    def update_chart(self):
        pass

    def retranslate(self):
        self.save_figure_action.setText(self.tr("Save Figure"))

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.StyleChange:
            setup_matplotlib()
            self.figure.clear()
            self.main_layout.removeWidget(self.canvas)
            self.canvas.setVisible(False)
            self.figure = plt.figure(figsize=self.figure.get_size_inches())
            self.canvas = FigureCanvas(self.figure)
            self.toolbar = NavigationToolbar(self.canvas, self)
            self.main_layout.addWidget(self.canvas, 0, 0)
            self.update_chart()

        elif event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()



import typing

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtCore, QtWidgets
from sklearn.decomposition import PCA

from ..model import GrainSizeDataset
from .BaseChart import BaseChart
from .config_matplotlib import highlight_color, normal_color


class PCAResultChart(BaseChart):
    def __init__(self, parent=None, figsize=(6, 5)):
        super().__init__(parent=parent, figsize=figsize)
        self.sample_axes = self.figure.add_subplot(2, 2, 1)
        self.shape_axes = self.figure.add_subplot(2, 2, 2)
        self.series_axes = self.figure.add_subplot(2, 1, 2)
        self.setWindowTitle(self.tr("PCA Chart"))
        self.__last_dataset = None # type: GrainSizeDataset

    def show_dataset(self, dataset: GrainSizeDataset):
        assert dataset is not None
        assert dataset.has_sample
        self.__last_dataset = dataset

        X = dataset.distribution_matrix
        pca = PCA()
        transformed = pca.fit_transform(X)
        # labels = [f"{v:0.4f}" for v in dataset.classes_μm]
        n_samples, n_features = X.shape
        cmap = plt.get_cmap()
        self.sample_axes.clear()
        self.shape_axes.clear()
        self.series_axes.clear()

        self.sample_axes.scatter(transformed[:, 0], transformed[:, 1], c=normal_color(), s=2.0, alpha=0.05)
        self.sample_axes.plot(pca.components_[0, :], pca.components_[1, :], color=normal_color(), lw=1.0, alpha=1.0)
        xi = np.argmax(pca.components_[0, :])
        yi = np.argmax(pca.components_[1, :])
        self.sample_axes.arrow(0, 0, pca.components_[0, xi], pca.components_[1, xi], color=cmap(0), width=0.001, alpha=1.0)
        self.sample_axes.text(pca.components_[0, xi], pca.components_[1, xi], f"{dataset.classes_μm[xi]: 0.4f}", color=cmap(0), ha='center', va='center')
        self.sample_axes.arrow(0, 0, pca.components_[0, yi], pca.components_[1, yi], color=cmap(1), width=0.001, alpha=1.0)
        self.sample_axes.text(pca.components_[0, yi], pca.components_[1, yi], f"{dataset.classes_μm[yi]: 0.4f}", color=cmap(1), ha='center', va='center')
        self.sample_axes.set_xlabel("PC1")
        self.sample_axes.set_ylabel("PC2")

        self.shape_axes.plot(dataset.classes_μm, pca.components_[0], color=cmap(0), label="PC1")
        self.shape_axes.plot(dataset.classes_μm, pca.components_[1], color=cmap(1), label="PC2")
        self.shape_axes.set_xscale("log")
        # plt.xticks([1e-1, 1e0, 1e1, 1e2, 1e3], ["0.1", "1", "10", "100", "1000"])
        self.shape_axes.set_xlabel("Grain size [μm]")
        self.shape_axes.set_ylabel("Transformed value")
        self.shape_axes.legend(loc="upper left")

        for i in range(2):
            self.series_axes.plot(transformed[:, i], color=cmap(i), label=f"PC{i+1} ({pca.explained_variance_ratio_[i]:0.2%})", lw=1.0, alpha=0.8)
        self.series_axes.set_xlabel("Sample index")
        self.series_axes.set_ylabel("Transformed value")
        self.series_axes.legend(loc="upper left")
        self.figure.tight_layout()
        self.canvas.draw()

    def update_chart(self):
        self.figure.clear()
        self.sample_axes = self.figure.add_subplot(2, 2, 1)
        self.shape_axes = self.figure.add_subplot(2, 2, 2)
        self.series_axes = self.figure.add_subplot(2, 1, 2)
        if self.__last_dataset is not None:
            self.show_dataset(self.__last_dataset)

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("PCA Chart"))



import typing

import numpy as np
from PySide6 import QtCore, QtWidgets
from scipy.cluster.hierarchy import dendrogram

from .BaseChart import BaseChart


class HierarchicalChart(BaseChart):
    def __init__(self, parent=None, figsize=(8, 6)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Hierarchical Clustering"))
        self.axes = self.figure.subplots()
        self.__linkage_matrix = None
        self.__p = None

    def show_result(self,
                    linkage_matrix: np.ndarray,
                    p=100):
        self.__linkage_matrix = linkage_matrix
        self.__p = p
        self.axes.clear()
        dendrogram(linkage_matrix,
                   no_labels=False,
                   p=p, truncate_mode='lastp',
                   show_contracted=True,
                   ax=self.axes)
        self.axes.set_xlabel("Sample count/index")
        self.axes.set_ylabel("Distance")
        self.figure.tight_layout()
        self.canvas.draw()

    def update_chart(self):
        self.figure.clear()
        self.axes = self.figure.subplots()
        if self.__linkage_matrix is not None:
            self.show_result(self.__linkage_matrix, self.__p)

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("Hierarchical Clustering"))



import typing

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FFMpegWriter, FuncAnimation, ImageMagickWriter
from PySide6 import QtCore, QtGui, QtWidgets

from ..statistic import convert_φ_to_μm
from ..udm import UDMResult
from .BaseChart import BaseChart
from .config_matplotlib import normal_color


def summarize(proportions: np.ndarray, components: np.ndarray, q=0.01):
    n_samples, n_components, n_classes = components.shape
    EM_mean = np.mean(np.expand_dims(proportions.squeeze(1), 2).repeat(n_classes, axis=2) * components, axis=0)
    EM_mean /= np.expand_dims(np.sum(EM_mean, axis=1), axis=1).repeat(n_classes, axis=1)
    EM_upper = np.zeros((n_components, n_classes))
    EM_lower = np.zeros((n_components, n_classes))
    for i in range(n_components):
        key = np.greater(proportions[:, 0, i], 1e-3)
        EM_upper[i] = np.quantile(components[:, i, :][key], q=1-q, axis=0)
        EM_lower[i] = np.quantile(components[:, i, :][key], q=q, axis=0)
    return EM_mean, EM_lower, EM_upper


class UDMResultChart(BaseChart):
    N_DISPLAY_SAMPLES = 200
    def __init__(self, parent=None, figsize=(4, 6)):
        super().__init__(parent=parent, figsize=figsize)
        # self.axes = self.figure.subplots()
        self.scale_menu = QtWidgets.QMenu(self.tr("Scale")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name) # type: QtGui.QAction
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)

        self.animated_action = QtGui.QAction(self.tr("Animated")) # type: QtGui.QAction
        self.animated_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.animated_action)
        self.animated_action.setCheckable(True)
        self.animated_action.setChecked(False)

        self.interval_menu = QtWidgets.QMenu(self.tr("Animation Interval")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.interval_menu)
        self.interval_group = QtGui.QActionGroup(self.interval_menu)
        self.interval_group.setExclusive(True)
        self.interval_actions = [] # type: list[QtGui.QAction]
        for interval, name in self.supported_intervals:
            interval_action = self.interval_group.addAction(name)
            interval_action.setCheckable(True)
            interval_action.triggered.connect(self.update_chart)
            self.interval_menu.addAction(interval_action)
            self.interval_actions.append(interval_action)
        self.interval_actions[3].setChecked(True)

        self.repeat_action = QtGui.QAction(self.tr("Repeat Animation")) # type: QtGui.QAction
        self.repeat_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.repeat_action)
        self.repeat_action.setCheckable(True)
        self.repeat_action.setChecked(False)

        self.save_animation_action = QtGui.QAction(self.tr("Save Animation")) # type: QtGui.QAction
        self.menu.addAction(self.save_animation_action)
        self.save_animation_action.triggered.connect(self.save_animation)

        self.__animation = None
        self.__last_result = None

        self.normal_msg = QtWidgets.QMessageBox(parent=self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

    @property
    def supported_scales(self) -> typing.List[typing.Tuple[str, str]]:
        scales = [("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear"))]
        return scales

    @property
    def supported_intervals(self) -> typing.List[typing.Tuple[int, str]]:
        intervals = [(5, self.tr("5 Milliseconds")),
                     (10, self.tr("10 Milliseconds")),
                     (20, self.tr("20 Milliseconds")),
                     (30, self.tr("30 Milliseconds")),
                     (60, self.tr("60 Milliseconds"))]
        return intervals

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def animated(self) -> bool:
        return self.animated_action.isChecked()

    @property
    def animation_interval(self) -> int:
        for i, interval_action in enumerate(self.interval_actions):
            if interval_action.isChecked():
                interval, name = self.supported_intervals[i]
                return interval

    @property
    def repeat_animation(self) -> bool:
        return self.repeat_action.isChecked()

    @property
    def transfer(self) -> typing.Callable:
        if self.scale == "log-linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)
        elif self.scale == "log":
            return lambda classes_φ: np.log(convert_φ_to_μm(classes_φ))
        elif self.scale == "phi":
            return lambda classes_φ: classes_φ
        elif self.scale == "linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size [μm]"
        elif self.scale == "log":
            return "Ln(grain size in μm)"
        elif self.scale == "phi":
            return "Grain size [φ]"
        elif self.scale == "linear":
            return "Grain size [μm]"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def show_chart(self, result: UDMResult):
        interval = max(1, result.n_samples//self.N_DISPLAY_SAMPLES)
        sample_indexes = np.linspace(1, result.n_samples, result.n_samples)
        iteration_indexes = np.linspace(1, result.n_iterations, result.n_iterations)
        classes = self.transfer(result.dataset.classes_φ)
        GSDs = result.dataset.distribution_matrix
        sum_loss_series = result.distribution_loss_series + result.component_loss_series

        GSDs_axes = self.figure.add_subplot(2, 2, 1)
        for sample in result.dataset.samples[::interval]:
            GSDs_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            GSDs_axes.set_xscale("log")
        GSDs_axes.set_xlim(classes[0], classes[-1])
        GSDs_axes.set_ylim(0.0, round(np.max(GSDs)*1.2, 2))
        GSDs_axes.set_xlabel(self.xlabel)
        GSDs_axes.set_ylabel(self.ylabel)
        GSDs_axes.set_title("GSDs")

        loss_axes = plt.subplot(2, 2, 2)
        loss_axes.plot(sum_loss_series, color=plt.get_cmap()(0), label="Sum")
        loss_axes.plot(result.distribution_loss_series, color=plt.get_cmap()(1), label="GSDs")
        loss_axes.plot(result.component_loss_series, color=plt.get_cmap()(2), label="Components")
        loss_axes.set_xlim(iteration_indexes[0], iteration_indexes[-1])
        loss_axes.set_xlabel("Iteration")
        loss_axes.set_ylabel("Loss")
        loss_axes.set_title("Loss variation")
        loss_axes.legend(loc="upper right", prop={"size": 8})

        component_axes = plt.subplot(2, 2, 3)
        mean, lower, upper = summarize(result.proportions, result.components, q=0.01)
        modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(mean[i]), mean[i].shape)]) for i in range(result.n_components)]
        # sort them by mode size
        modes.sort(key=lambda x: x[1])
        for i_c, (i, _) in enumerate(modes):
            component_axes.plot(classes, mean[i], c=plt.get_cmap()(i_c), zorder=20+i_c)
            component_axes.fill_between(
                classes, lower[i], upper[i],
                color=plt.get_cmap()(i_c),
                alpha=0.2, lw=0.02, zorder=10+i_c)
        if self.xlog:
            component_axes.set_xscale("log")
        component_axes.set_xlim(classes[0], classes[-1])
        component_axes.set_ylim(0.0, round(np.max(mean)*1.2, 2))
        component_axes.set_xlabel(self.xlabel)
        component_axes.set_ylabel(self.ylabel)
        component_axes.set_title("Components")

        proportion_axes = plt.subplot(2, 2, 4)
        bottom = np.zeros(result.n_samples)
        for i_em, (i, _) in enumerate(modes):
            proportion_axes.bar(
                sample_indexes[::interval],
                result.proportions[:, 0, i][::interval],
                bottom=bottom[::interval],
                width=interval, color=plt.get_cmap()(i_em))
            bottom += result.proportions[:, 0, i]
        proportion_axes.set_xlim(sample_indexes[0], sample_indexes[-1])
        proportion_axes.set_ylim(0.0, 1.0)
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")
        self.figure.tight_layout()
        self.canvas.draw()

    def show_animation(self, result: UDMResult):
        interval = max(1, result.n_samples // self.N_DISPLAY_SAMPLES)
        sample_indexes = np.linspace(1, result.n_samples, result.n_samples)
        iteration_indexes = np.linspace(1, result.n_iterations, result.n_iterations)
        classes = self.transfer(result.dataset.classes_φ)
        GSDs = result.dataset.distribution_matrix
        sum_loss_series = result.distribution_loss_series + result.component_loss_series
        distances = np.array([result.distribution_loss_series, result.component_loss_series, sum_loss_series])
        min_distance, max_distance = np.min(distances), np.max(distances)

        GSDs_axes = self.figure.add_subplot(2, 2, 1)
        for sample in result.dataset.samples[::interval]:
            GSDs_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            GSDs_axes.set_xscale("log")
        GSDs_axes.set_xlim(classes[0], classes[-1])
        GSDs_axes.set_ylim(0.0, round(np.max(GSDs)*1.2, 2))
        GSDs_axes.set_xlabel(self.xlabel)
        GSDs_axes.set_ylabel(self.ylabel)
        GSDs_axes.set_title("GSDs")

        loss_axes = plt.subplot(2, 2, 2)
        sum_loss_line = loss_axes.plot(sum_loss_series, color=plt.get_cmap()(0), label="Sum")[0]
        distribution_loss_line = loss_axes.plot(result.distribution_loss_series, color=plt.get_cmap()(1), label="GSDs")[0]
        component_loss_line = loss_axes.plot(result.component_loss_series, color=plt.get_cmap()(2), label="Components")[0]
        loss_axes.set_xlim(iteration_indexes[0], iteration_indexes[-1])
        loss_axes.set_xlabel("Iteration")
        loss_axes.set_ylabel("Loss")
        loss_axes.set_title("Loss variation")
        loss_axes.legend(
            handles=[sum_loss_line, distribution_loss_line, component_loss_line],
            labels=["Sum", "GSDs", "Components"],
            loc="upper right", prop={"size": 8})
        component_axes = plt.subplot(2, 2, 3)
        mean, lower, upper = summarize(result.proportions, result.components, q=0.01)
        modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(mean[i]), mean[i].shape)]) for i in range(result.n_components)]
        # sort them by mode size
        modes.sort(key=lambda x: x[1])
        if self.xlog:
            component_axes.set_xscale("log")
        component_axes.set_xlim(classes[0], classes[-1])
        component_axes.set_ylim(0.0, round(np.max(mean)*1.2, 2))
        component_axes.set_xlabel(self.xlabel)
        component_axes.set_ylabel(self.ylabel)
        component_axes.set_title("Components")

        proportion_axes = plt.subplot(2, 2, 4)
        proportion_axes.set_xlim(sample_indexes[0], sample_indexes[-1])
        proportion_axes.set_ylim(0.0, 1.0)
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")
        # self.figure.tight_layout()
        # self.canvas.draw()

        def init():
            self.iteration_line = loss_axes.plot([1, 1], [min_distance, max_distance], c=normal_color())[0]
            self.component_curves = []
            self.component_shadows = []
            mean, lower, upper = summarize(result.proportions, result.components, q=0.01)
            for i_c, (i, _) in enumerate(modes):
                component_curve = component_axes.plot(classes, mean[i], c=plt.get_cmap()(i_c), zorder=20+i_c)[0]
                component_shadow = component_axes.fill_between(
                    classes, lower[i], upper[i],
                    color=plt.get_cmap()(i_c),
                    alpha=0.2, lw=0.02, zorder=10+i_c)
                self.component_curves.append(component_curve)
                self.component_shadows.append(component_shadow)

            bottom = np.zeros(result.n_samples)
            self.proportion_bars = []
            self.patches = []
            for i_em, (index, _) in enumerate(modes):
                bar = proportion_axes.bar(
                    sample_indexes[::interval],
                    result.proportions[:, 0, index][::interval],
                    bottom=bottom[::interval],
                    width=interval, color=plt.get_cmap()(i_em))
                self.proportion_bars.append(bar)
                self.patches.extend(bar.patches)
                bottom += result.proportions[:, 0, index]
            return self.iteration_line, *(self.component_curves + self.component_shadows + self.patches)

        def animate(args: typing.Tuple[int, UDMResult]):
            iteration, current = args
            mean, lower, upper = summarize(current.proportions, current.components, q=0.01)
            self.iteration_line.set_xdata([iteration, iteration])
            for i_em, (index, _) in enumerate(modes):
                self.component_curves[i_em].set_ydata(mean[index])
                verts_lower = np.concatenate([np.expand_dims(classes, axis=1), np.expand_dims(lower[i_em], axis=1)], axis=1)
                verts_upper = np.concatenate([np.expand_dims(classes[::-1], axis=1), np.expand_dims(upper[i_em][::-1], axis=1)], axis=1)
                verts = np.concatenate([verts_lower, verts_upper], axis=0)
                self.component_shadows[i_em].set_verts([verts])
            bottom = np.zeros(current.n_samples)
            for i_em, (index, _) in enumerate(modes):
                for rect, height, y in zip(self.proportion_bars[i_em].patches, current.proportions[:, 0, index][::interval], bottom[::interval]):
                    rect.set_height(height)
                    rect.set_y(y)
                bottom += current.proportions[:, 0, index]
            return self.iteration_line, *(self.component_curves + self.component_shadows + self.patches)

        self.__animation = FuncAnimation(
            self.figure, animate, init_func=init,
            frames=enumerate(result.history), interval=self.animation_interval,
            blit=True, repeat=self.repeat_animation,
            repeat_delay=3.0, save_count=result.n_iterations)

    def show_result(self, result: UDMResult):
        self.__last_result = result
        self.figure.clear()
        if self.__animation is not None:
            self.__animation._stop()
            self.__animation = None
        if self.animated:
            self.show_animation(result)
        else:
            self.show_chart(result)

    def update_chart(self):
        if self.__last_result is not None:
            self.show_result(self.__last_result)

    def save_animation(self):
        if self.__last_result is not None:
            filename, format_str  = self.file_dialog.getSaveFileName(
                self, self.tr("Choose a filename to save the animation of this UDM result"),
                None, "MPEG-4 Video File (*.mp4);;Graphics Interchange Format (*.gif)")
            if filename is None or filename == "":
                return
            progress = QtWidgets.QProgressDialog(self)
            progress.setWindowTitle("QGrain")
            progress.setRange(0, 100)
            progress.setLabelText(self.tr("Saving Animation [{0} Frames]").format(self.__last_result.n_iterations))
            canceled = False
            def save_callback(i, n):
                if progress.wasCanceled():
                    nonlocal canceled
                    canceled = True
                    raise StopIteration()
                progress.setValue((i+1)/n*100)
                QtCore.QCoreApplication.processEvents()
            self.animated_action.setChecked(True)
            self.update_chart()
            # plt.rcParams["savefig.dpi"] = 120.0
            if "*.gif" in format_str:
                if not ImageMagickWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("ImageMagick is not installed, please download and install it from its offical website (https://imagemagick.org/index.php)."))
                    self.normal_msg.exec_()
                else:
                    self.__animation.save(filename, writer="imagemagick", fps=30, progress_callback=save_callback)
            elif "*.mp4" in format_str:
                if not FFMpegWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("FFMpeg is not installed, please download and install it from its offical website (https://ffmpeg.org/)."))
                    self.normal_msg.exec_()
                else:
                    self.__animation.save(filename, writer="ffmpeg", fps=30, progress_callback=save_callback)
            # plt.rcParams["savefig.dpi"] = 300.0
            if not canceled:
                progress.setValue(100)

    def retranslate(self):
        super().retranslate()
        self.scale_menu.setTitle(self.tr("Scale"))
        for action, (key, name) in zip(self.scale_actions, self.supported_scales):
            action.setText(name)
        self.animated_action.setText(self.tr("Animated"))
        self.interval_menu.setTitle(self.tr("Animation Interval"))
        for action, (interval, name) in zip(self.interval_actions, self.supported_intervals):
            action.setText(name)
        self.repeat_action.setText(self.tr("Repeat Animation"))
        self.save_animation_action.setText(self.tr("Save Animation"))



import csv
import logging
import os
import typing
from enum import Enum, unique

import numpy as np
import openpyxl
import xlrd

from ..model import GrainSizeDataset


class DataLayoutError(Exception):
    """Raises while the data layout settings are invalid."""
    pass


class DataLayoutSetting:
    """
    The class to represent the layout setting of raw data file.
    All types of raw data files can be regarded as the table(s) that contains rows and columns.
    For the file which contains grain size distributions, it should be the following format:
        * The first valid row should be the headers (i.e. the grain size classes).
        * The following valid rows should be the distributions of samples corresponding to the grain size classes.
        * The first valid column shoud be the names of samples.
    To make it more flexible, we use this setting to control the data loader.
    """
    def __init__(self, classes_row=0, sample_names_column=0,
                 distribution_start_row=1, distribution_start_column=1):
        # make sure the types are int in other codes
        assert isinstance(classes_row, int)
        assert isinstance(sample_names_column, int)
        assert isinstance(distribution_start_row, int)
        assert isinstance(distribution_start_column, int)
        # handle these errors at front-end
        if classes_row < 0 or \
            sample_names_column < 0 or \
            distribution_start_row < 0 or \
            distribution_start_column < 0:
            raise DataLayoutError("Index must be non-negative.")
        if classes_row >= distribution_start_row:
            raise DataLayoutError("The start row index of distribution must be greater than the row index of classes.")
        if sample_names_column >= distribution_start_column:
            raise DataLayoutError("The start column index of distribution must be greater than the column index of sample names.")
        self.__classes_row = classes_row
        self.__sample_names_column = sample_names_column
        self.__distribution_start_row = distribution_start_row
        self.__distribution_start_column = distribution_start_column

    @property
    def classes_row(self) -> int:
        return self.__classes_row

    @property
    def sample_names_column(self) -> int:
        return self.__sample_names_column

    @property
    def distribution_start_row(self) -> int:
        return self.__distribution_start_row

    @property
    def distribution_start_column(self) -> int:
        return self.__distribution_start_column


@unique
class ReadFileType(Enum):
    XLS = 0
    XLSX = 1
    CSV = 2


def get_type_by_name(filename: str):
    _, extension = os.path.splitext(filename)
    if extension == ".csv":
        return ReadFileType.CSV
    elif extension == ".xls":
        return ReadFileType.XLS
    elif extension == ".xlsx":
        return ReadFileType.XLSX
    else:
        raise NotImplementedError(extension)


def _get_raw_table(
    filename: str, file_type: ReadFileType, sheet_index: int,
    progress_callback: typing.Callable = None, logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug(f"Try to load raw data table from [{file_type.name}] file: {filename}.")
    if file_type == ReadFileType.CSV:
        with open(filename, encoding="utf-8") as f:
            reader = csv.reader(f)
            raw_table = []
            for i, row_values in enumerate(reader):
                raw_table.append(row_values)
                if progress_callback is not None:
                    progress_callback(i / reader.line_num)

    elif file_type == ReadFileType.XLS:
        workbook = xlrd.open_workbook(filename) # type: xlrd.Book
        sheet = workbook.sheet_by_index(sheet_index)
        raw_table = []
        for row in range(sheet.nrows):
            raw_table.append(sheet.row_values(row))
            if progress_callback is not None:
                progress_callback(row / sheet.nrows)

    elif file_type == ReadFileType.XLSX:
        workbook = openpyxl.load_workbook(filename, read_only=True, data_only=True) # type: openpyxl.Workbook
        sheet = workbook[workbook.sheetnames[sheet_index]]
        raw_table = []
        for i, row_values in enumerate(sheet.values):
            raw_table.append(row_values)
            if progress_callback is not None:
                progress_callback(i / sheet.max_row)
    else:
        raise NotImplementedError(file_type)
    logger.debug("Raw data table has been loaded from this file.")
    return raw_table


def load_dataset(filename: str,
                 sheet_index: int = 0,
                 layout: DataLayoutSetting = None,
                 progress_callback: typing.Callable = None,
                 logger: logging.Logger = None):
    """
    Try to load the grain size dataset from the file.

    ## Parameters

    filename: `str`
        The file path which contains the extension.
    sheet_index: `int`
        If it is an Excel file (`*.xls` or `*.xlsx`), please indicate the sheet index, otherwise it will use the first sheet.
    layout: `DataLayoutSetting`
        The layout of the grain size data stored in the sheet.
    logger: `logging.Logger`
        The logger which will be used to log the information of loading.

    ## Returns

    result: `None` or `GrainSizeDataset`
        If there is any exception raised, it will return `None`.

    """
    assert isinstance(filename, str)
    assert len(filename) != 0
    assert isinstance(sheet_index, int)
    assert sheet_index >= 0
    if layout is None:
        layout = DataLayoutSetting()
    else:
        assert isinstance(layout, DataLayoutSetting)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    file_type = get_type_by_name(filename)
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress*0.5)
    else:
        _callback = None
    raw_table = _get_raw_table(filename, file_type, sheet_index, progress_callback=_callback)
    try:
        class_values = raw_table[layout.classes_row][layout.distribution_start_column:]
        classes_μm = np.array(class_values, dtype=np.float64)
    except IndexError as e:
        logger.exception(f"The row index of classes [{layout.classes_row}] or start column index [{layout.distribution_start_column}] of distribution is out of range, please check.", stack_info=True)
        return
    except ValueError as e:
        logger.exception(f"Can not convert the values of classes to a numerical array, it may contains invalid values (e.g. text).", stack_info=True)
        return
    logger.debug(f"Grain size classes in μm: [{','.join([f'{x: 0.4f}' for x in classes_μm])}].")
    try:
        GrainSizeDataset.validate_classes_μm(classes_μm)
    except Exception as e:
        logger.exception(f"The array of grain size classes is invalid, please check.", stack_info=True)
        return
    logger.debug("The array of grain size classes is valid.")

    try:
        names = []
        distributions = []
        for current_row, row_values in enumerate(raw_table[layout.distribution_start_row:], layout.distribution_start_row+1):
            # check if it's a empty row, i.e. the values all are empty string
            is_empty_row = True
            for distribution_value in row_values[layout.distribution_start_column:]:
                if distribution_value != "" and distribution_value is not None:
                    is_empty_row = False
                    break
            # if it's a empty row, jump this row to process the next one
            if is_empty_row:
                logger.warning(f"Row [{current_row}] is empty, skip to next row.")
                continue

            sample_name = row_values[layout.sample_names_column]
            # logger.debug(f"Processing the row [{current_row}].")
            if sample_name is None:
                sample_name = "NONE"
                logger.warning(f"The sample name is invalid, use 'NONE' instead.")
            # users may use pure number as the sample name
            elif not isinstance(sample_name, str):
                sample_name = str(sample_name)
                logger.warning(f"The sample name is not text (may be a number), convert it to text.")
            elif sample_name == "":
                sample_name = "EMPTY"
                logger.warning(f"The sample name is a empty text, use 'EMPTY' instead.")
            try:
                distribution = np.array(row_values[layout.distribution_start_column:], dtype=np.float64)
            except ValueError as e:
                logger.exception(f"Can not convert the distribution values at row [{current_row}] to a numerical array, it may contains invalid values (e.g. text or empty cell).", stack_info=True)
                return
            try:
                GrainSizeDataset.validate_distribution(distribution)
            except Exception as e:
                logger.exception(f"The distribution array of sample [{sample_name}] at row [{current_row}] is invalid.", stack_info=True)
                return
            names.append(sample_name)
            distributions.append(distribution)
            # logger.debug(f"The validation of sample [{sample_name}] at row [{current_row}] is passed.")
            if progress_callback is not None:
                progress = current_row / len(raw_table) * 0.5 + 0.5
                progress_callback(progress)
    except IndexError as e:
        logger.exception("At least one row or column index of data layout is out of range, please check.", stack_info=True)
        return
    try:
        dataset = GrainSizeDataset()
        dataset.add_batch(classes_μm, names, distributions, need_validation=False)
    except Exception as e:
        logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
        return
    logger.info("This dataset has been loaded successfully.")
    return dataset



__all__ = ["EMMAAnalyzer"]

import logging
import os
import pickle
import typing

import numpy as np
from PySide6 import QtCore, QtWidgets

from ..chart.EMMAResultChart import EMMAResultChart
from ..emma import EMMAResolver, EMMAResult, KernelType
from ..io import save_emma
from ..model import GrainSizeDataset
from .EMMASettingDialog import EMMASettingDialog
from .ParameterEditor import ParameterEditor


class EMMAAnalyzer(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.EMMAAnalyzer")
    SUPPORT_KERNELS = (
        KernelType.Nonparametric,
        KernelType.Normal,
        KernelType.SkewNormal,
        KernelType.Weibull,
        KernelType.GeneralWeibull)
    def __init__(self, setting_dialog: EMMASettingDialog,
                 parameter_editor: ParameterEditor, parent=None):
        super().__init__(parent=parent)
        assert setting_dialog is not None
        assert parameter_editor is not None
        self.setting_dialog = setting_dialog
        self.parameter_editor = parameter_editor
        self.init_ui()
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.__dataset = None # type: GrainSizeDataset
        self.__result_list = [] # type: list[EMMAResult]
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

    def init_ui(self):
        self.setWindowTitle(self.tr("EMMA Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        # Control group
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.kernel_label = QtWidgets.QLabel(self.tr("Kernel Type"))
        self.kernel_combo_box = QtWidgets.QComboBox()
        self.kernel_combo_box.addItems([kernel_type.value for kernel_type in self.SUPPORT_KERNELS])
        self.kernel_combo_box.setCurrentIndex(1)
        self.control_layout.addWidget(self.kernel_label, 0, 0)
        self.control_layout.addWidget(self.kernel_combo_box, 0, 1)
        self.n_members_label = QtWidgets.QLabel(self.tr("Number of End Members"))
        self.n_members_input = QtWidgets.QSpinBox()
        self.n_members_input.setRange(1, 12)
        self.n_members_input.setValue(3)
        self.control_layout.addWidget(self.n_members_label, 1, 0)
        self.control_layout.addWidget(self.n_members_input, 1, 1)
        self.update_EMs_checkbox = QtWidgets.QCheckBox(self.tr("Update End Members"))
        self.update_EMs_checkbox.setChecked(True)
        self.control_layout.addWidget(self.update_EMs_checkbox, 2, 0, 1, 2)
        self.try_fit_button = QtWidgets.QPushButton(self.tr("Try Fit"))
        self.try_fit_button.clicked.connect(self.on_try_fit_clicked)
        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button = QtWidgets.QPushButton(self.tr("Edit Parameters"))
        self.edit_parameter_button.clicked.connect(self.on_edit_parameter_clicked)
        self.edit_parameter_button.setEnabled(False)
        self.control_layout.addWidget(self.try_fit_button, 3, 0)
        self.control_layout.addWidget(self.edit_parameter_button, 3, 1)

        # Result group
        self.result_group = QtWidgets.QGroupBox(self.tr("Result"))
        self.result_group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.result_layout = QtWidgets.QGridLayout(self.result_group)
        self.result_list_widget = QtWidgets.QListWidget()
        self.result_layout.addWidget(self.result_list_widget, 0, 0, 1, 2)
        self.remove_button = QtWidgets.QPushButton(self.tr("Remove"))
        self.remove_button.clicked.connect(self.on_remove_clicked)
        self.show_button = QtWidgets.QPushButton(self.tr("Show"))
        self.show_button.clicked.connect(self.on_show_clicked)
        self.remove_button.setEnabled(False)
        self.show_button.setEnabled(False)
        self.result_layout.addWidget(self.remove_button, 1, 0)
        self.result_layout.addWidget(self.show_button, 1, 1)

        self.chart_group = QtWidgets.QGroupBox(self.tr("Chart"))
        self.chart_group.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.chart_layout = QtWidgets.QGridLayout(self.chart_group)
        self.result_chart = EMMAResultChart()
        self.chart_layout.addWidget(self.result_chart)

        self.splitter_1 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.result_group)
        self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.chart_group)
        self.main_layout.addWidget(self.splitter_2)

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def kernel_type(self) -> KernelType:
        kernel_type = self.SUPPORT_KERNELS[self.kernel_combo_box.currentIndex()]
        return kernel_type

    @property
    def n_members(self):
        return self.n_members_input.value()

    @property
    def n_results(self) -> int:
        return len(self.__result_list)

    @property
    def selected_index(self):
        indexes = self.result_list_widget.selectedIndexes()
        if len(indexes) == 0:
            return 0
        else:
            return indexes[0].row()

    @property
    def selected_result(self):
        if self.n_results > 0:
            return self.__result_list[self.selected_index]

    def on_dataset_loaded(self, dataset: GrainSizeDataset):
        self.__dataset = dataset
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)

    def on_edit_parameter_clicked(self):
        if self.__dataset is not None:
            target = np.mean(self.__dataset.distribution_matrix, axis=0)
            target = target / np.sum(target)
            self.parameter_editor.setup_target(self.__dataset.classes_μm, target)
        self.parameter_editor.show()

    def on_try_fit_clicked(self):
        if self.__dataset is None:
            self.logger.error("Dataset has not been loaded.")
            self.show_error(self.tr("Dataset has not been loaded."))
            return
        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button.setEnabled(False)
        resolver = EMMAResolver()
        resolver_setting = self.setting_dialog.setting
        update_end_members = self.update_EMs_checkbox.isChecked()
        if self.parameter_editor.parameter_enabled:
            self.logger.info("The parameters in Parameter Editor are enabled. They will be used preferentially!")
            kernel_type = KernelType.__members__[self.parameter_editor.distribution_type.name]
            parameters = self.parameter_editor.parameters[:-1, :].astype(np.float32)
            self.logger.debug(f"Try to perform the EMMA algorithm. Kernel type: {kernel_type.name}. Number of end members: {self.parameter_editor.n_components}. Initial parameters: {parameters}. Algorithm settings: {resolver_setting}. Update the end members: {update_end_members}.")
            result = resolver.try_fit(
                self.__dataset, kernel_type,
                self.parameter_editor.n_components,
                resolver_setting=resolver_setting,
                parameters=parameters,
                update_end_members=update_end_members)
        else:
            self.logger.debug(f"Try to perform the EMMA algorithm. Kernel type: {self.kernel_type.name}. Number of end members: {self.n_members}. Algorithm settings: {resolver_setting}. Update the end members: {update_end_members}.")
            result = resolver.try_fit(
                self.__dataset, self.kernel_type,
                self.n_members,
                resolver_setting=resolver_setting,
                update_end_members=update_end_members)
        self.add_results([result])
        self.result_chart.show_result(result)
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)

    def get_result_name(self, result: EMMAResult):
        if self.update_EMs_checkbox.isChecked():
            fixed_str = " "
        else:
            fixed_str = " Fixed "
        return f"{result.n_members}{fixed_str}{result.kernel_type.value}"

    def add_results(self, results: typing.List[EMMAResult]):
        if self.n_results == 0:
            self.remove_button.setEnabled(True)
            self.show_button.setEnabled(True)

        self.__result_list.extend(results)
        self.result_list_widget.addItems([self.get_result_name(result) for result in results])

    def on_remove_clicked(self):
        if self.n_results == 0:
            return
        else:
            index = self.selected_index
            self.__result_list.pop(index)
            self.result_list_widget.takeItem(index)

        if self.n_results == 0:
            self.remove_button.setEnabled(False)
            self.show_button.setEnabled(False)

    def on_show_clicked(self):
        result = self.selected_result
        if result is not None:
            self.result_chart.show_result(result)

    def load_result(self):
        filename, _  = self.file_dialog.getOpenFileName(
            self, self.tr("Choose the file which stores the dumped EMMA result"),
            None, "Dumped EMMA Result (*.emma)")
        if filename is None or filename == "":
            return
        with open(filename, "rb") as f:
            result = pickle.load(f)
            if isinstance(result, EMMAResult):
                self.add_results([result])
                self.logger.info(f"The dumped EMMA result has been loaded.")
            else:
                self.logger.error("The binary file is invalid (i.e., the object in it is not the EMMA result).")
                self.show_error(self.tr("The binary file is invalid (i.e., the object in it is not the EMMA result)."))
                return

    def save_selected_result(self):
        if self.n_results == 0:
            self.logger.error("There is no EMMA result.")
            self.show_error(self.tr("There is no EMMA result."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the selected EMMA result"),
            None, "Microsoft Excel (*.xlsx);;Dumped EMMA Result (*.emma)")
        if filename is None or filename == "":
            return
        try:
            # Excel
            if filename[-4:] == "xlsx":
                result = self.selected_result
                progress_dialog = QtWidgets.QProgressDialog(
                    self.tr("Saving the EMMA result..."), self.tr("Cancel"),
                    0, 100, self)
                progress_dialog.setWindowTitle("QGrain")
                progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
                def callback(progress: float):
                    if progress_dialog.wasCanceled():
                        raise StopIteration()
                    progress_dialog.setValue(int(progress*100))
                    QtCore.QCoreApplication.processEvents()
                save_emma(result, filename, progress_callback=callback, logger=self.logger)
            # Binary File
            else:
                with open(filename, "wb") as f:
                    pickle.dump(self.selected_result, f)
                    self.logger.info("The selected EMMA result has been dumped.")
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("EMMA Analyzer"))
        self.control_group.setTitle(self.tr("Control"))
        self.kernel_label.setText(self.tr("Kernel Type"))
        self.n_members_label.setText(self.tr("Number of End Members"))
        self.update_EMs_checkbox.setText(self.tr("Update End Members"))
        self.try_fit_button.setText(self.tr("Try Fit"))
        self.edit_parameter_button.setText(self.tr("Edit Parameters"))
        self.result_group.setTitle(self.tr("Result"))
        self.remove_button.setText(self.tr("Remove"))
        self.show_button.setText(self.tr("Show"))
        self.chart_group.setTitle(self.tr("Chart"))



import numpy as np
import torch


def log10MSE_distance(values: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    return torch.log10(torch.mean(torch.square(values - targets)))

def MSE_distance(values: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    return torch.mean(torch.square(values - targets))

def p_norm(values: torch.Tensor, targets: torch.Tensor, p=2) -> torch.Tensor:
    return torch.sum(torch.abs(values - targets) ** p) ** (1 / p)

def cosine_distance(values: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    cosine = torch.sum(values * targets) / (torch.sqrt(torch.sum(torch.square(values))) * torch.sqrt(torch.sum(torch.square(targets))))
    return torch.abs(cosine)

def angular_distance(values: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    cosine = cosine_distance(values, targets)
    angular = 2 * torch.arccos(cosine) / np.pi
    return angular

def get_distance_func_by_name(distance: str):
    if distance[-4:] == "norm":
        p = int(distance[0])
        return lambda x, y: p_norm(x, y, p)
    elif distance == "MSE":
        return lambda x, y: MSE_distance(x, y)
    elif distance == "log10MSE":
        return lambda x, y: log10MSE_distance(x, y)
    elif distance == "angular":
        return lambda x, y: angular_distance(x, y)
    else:
        raise NotImplementedError(distance)



import typing

import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import find_peaks


def get_mode(classes_μm: np.ndarray,
             classes_φ: np.ndarray,
             frequency: np.ndarray,
             is_geometric: bool = False) -> float:
    """
    Get the mode size of the grain size distribution.

    ## Parameters

    classes_μm: `np.ndarray`
        The grain size classes in μm.
    classes_φ: `np.ndarray`
        The grain size classes in φ.
    frequency: `np.ndarray`
        The frequency values of grain size classes. Note that the sum of frequencies should be `1.0`.
    is_geometric: `bool` (default `False`)
        If `True` the mode size is in μm, else it's in φ.

    ## Returns

    mode: `float`
        The mode size of a grain size distribution.

    """
    max_pos = np.unravel_index(np.argmax(frequency), frequency.shape)
    if is_geometric:
        return classes_μm[max_pos]
    else:
        return classes_φ[max_pos]

def get_modes(classes_μm: np.ndarray,
              classes_φ: np.ndarray,
              frequency: np.ndarray,
              is_geometric=False, trace=0.01) \
        -> typing.Tuple[typing.Tuple[float], typing.Tuple[float]]:
    """
    Get the mode sizes (peaks) of a grain size distribution. The number may be greater than 1.

    ## Parameters

    classes_μm: `np.ndarray`
        The grain size classes in μm.
    classes_φ: `np.ndarray`
        The grain size classes in φ.
    frequency: `np.ndarray`
        The frequency values of grain size classes. Note that the sum of frequencies should be `1.0`.
    is_geometric: `bool` (default `False`)
        If `True` the mode size is in μm, else it's in φ.
    trace: `float` (default `0.01`, `1%`)
        If the peak's frequency is less than `trace`, it will be ignored.

    ## Returns

    modes: `tuple`
        The mode sizes of a grain size distribution.
    frequencies: `tuple`
        The frequencies of corresponding mode sizes.

    """
    peaks, _ = find_peaks(frequency)
    position = classes_μm[peaks] if is_geometric else classes_φ[peaks]
    values = frequency[peaks]
    not_trace = np.greater_equal(values, trace)
    return tuple(position[not_trace]), tuple(values[not_trace])

def get_cumulative_frequency(frequency: np.ndarray, expand=False):
    """
    Transform the frequency distribution to cumulative frequency.

    ## Parameters

    frequency: `np.ndarray`
        The frequency values of grain size classes. Note that the sum of frequencies should be `1.0`.
    expand: `bool` (default `False`)
        If `True`, `0.0` and `1.0` will be added at the head and tail, respectively.

    ## Returns

    culmulative_frequency: `np.ndarray`
        A numpy array contains the culmulative frequencies.

    """
    cumulative_frequency = []
    cumulative = 0.0
    if expand:
        cumulative_frequency.append(0.0)
    for i, f in enumerate(frequency):
        cumulative += f
        cumulative_frequency.append(cumulative)
    if expand:
        cumulative_frequency.append(1.0)
    return np.array(cumulative_frequency)

def get_reverse_phi_ppf(classes_φ: np.ndarray, frequency: np.ndarray) -> interp1d:
    """
    Get the reversed percent point function (PPF) of a grain size distribution.
    Because the grain size classes in φ is isometric, it's more suitable for interpolation.
    And because coarser particles have smaller φ values, the PPF is reversed.

    ## Parameters

    classes_φ: `np.ndarray`
        The grain size classes in φ.
    frequency: `np.ndarray`
        The frequency values of grain size classes. Note that the sum of frequencies should be `1.0`.

    ## Returns

    ppf: `scipy.interpolate.interp1d`
        A `interp1d` object which is callable and could be regarded as a PPF function.

    """

    interval = np.mean(classes_φ[1:] - classes_φ[:-1])
    expand_classes = np.linspace(classes_φ[0] - interval, classes_φ[-1]+interval, len(classes_φ)+2)
    cumulative_frequency = get_cumulative_frequency(frequency, expand=True)
    cumulative_frequency = np.array(cumulative_frequency)
    ppf = interp1d(cumulative_frequency, expand_classes, kind="slinear")
    return ppf

def get_interval_φ(classes_φ: np.ndarray)-> float:
    """
    Get the interval of grain size classes.

    ## Parameters

    classes_φ: `np.ndarray`
        The grain size classes in φ.

    ## Returns

    interval_φ: `float`
        The interval in φ.

    """

    return abs((classes_φ[0]-classes_φ[-1]) / (len(classes_φ)-1))

def convert_μm_to_φ(size_μm: typing.Union[float, np.ndarray]):
    """
    Convert the grain size from μm space to φ space.

    ## Parameters

    size_μm: `float` or `np.ndarray`
        The grain size in μm.

    ## Returns

    size_φ: `np.ndarray`
        The grain size in φ.

    """

    return -np.log2(size_μm / 1000.0)

def convert_φ_to_μm(size_φ: np.ndarray):
    """
    Convert the grain size from φ space to μm space.

    ## Parameters

    size_φ: `float` or `np.ndarray`
        The grain size in φ.

    ## Returns

    size_μm: `np.ndarray`
        The grain size in μm.

    """

    return np.exp2(-size_φ) * 1000.0



import logging
import multiprocessing
import time
import traceback
import typing

import numpy as np
from PySide6.QtCore import QObject, Qt, QThread, Signal, Slot
from scipy.optimize import OptimizeResult, basinhopping, minimize

from ..model import GrainSizeDataset, GrainSizeSample
from ._distance import get_distance_function
from ._distribution import DISTRIBUTION_CLASS_MAP, DistributionType
from ._result import SSUResult
from ._task import SSUTask

# "cosine" has problem
built_in_distances = (
    "1-norm", "2-norm", "3-norm", "4-norm",
    "MSE", "log10MSE", "angular")

def check_distance(distance):
    assert isinstance(distance, str)
    in_set = False
    for d in built_in_distances:
        if distance == d:
            in_set = True
            break
    assert in_set


built_in_minimizers = (
    "Nelder-Mead", "Powell", "CG", "BFGS",
    "L-BFGS-B", "TNC", "SLSQP")

def check_minimizer(minimizer):
    assert isinstance(minimizer, str)
    in_set = False
    for m in built_in_minimizers:
        if minimizer == m:
            in_set = True
            break
    assert in_set


class SSUAlgorithmSetting:
    def __init__(self,
                 distance: str="log10MSE",
                 minimizer: str="SLSQP",
                 try_GO: bool=False,
                 GO_max_niter: int=100,
                 GO_success_niter: int=5, # GO - Global Optimization
                 GO_step: float=0.1,
                 GO_minimizer_max_niter: int=500,
                 LO_max_niter: int=1000):
        # validation
        check_distance(distance)
        check_minimizer(minimizer)
        assert isinstance(try_GO, bool)
        assert isinstance(GO_max_niter, int)
        assert isinstance(GO_success_niter, int)
        assert isinstance(GO_step, (int, float))
        assert isinstance(GO_minimizer_max_niter, int)
        assert isinstance(LO_max_niter, int)
        assert GO_max_niter > 0
        assert GO_success_niter > 0
        assert GO_step > 0.0
        assert GO_minimizer_max_niter > 0
        assert LO_max_niter > 0

        self.distance = distance
        self.minimizer = minimizer
        self.try_GO = try_GO
        self.GO_max_niter = GO_max_niter
        self.GO_success_niter = GO_success_niter
        self.GO_step = GO_step
        self.GO_minimizer_max_niter = GO_minimizer_max_niter
        self.LO_max_niter = LO_max_niter

    def __str__(self) -> str:
        return self.__dict__.__str__()


class BasicResolver:
    def __init__(self, hooks=None):
        if hooks is not None:
            for name, func in hooks.items():
                self.__setattr__(name, func)

    # hooks
    def on_fitting_started(self):
        pass

    def on_fitting_finished(self):
        pass

    def on_global_fitting_failed(self, algorithm_result: OptimizeResult):
        pass

    def on_global_fitting_succeeded(self, algorithm_result: OptimizeResult):
        pass

    def on_final_fitting_failed(self, algorithm_result: OptimizeResult):
        pass

    def on_exception_raised_while_fitting(self, exception: Exception):
        pass

    def local_iteration_callback(self, fitted_params: typing.Iterable[float]):
        pass

    def global_iteration_callback(self, fitted_params: typing.Iterable[float], function_value: float, accept: bool):
        pass

    def on_fitting_succeeded(self, algorithm_result: OptimizeResult, fitting_result: SSUResult):
        pass

    def get_weights(self, classes_φ: np.ndarray, distribution: np.ndarray):
        from scipy.signal import find_peaks
        peaks, info = find_peaks(distribution)
        non_zeros = np.argwhere(distribution)
        start, end = non_zeros[0].max(), non_zeros[-1].max()
        weights = np.ones_like(distribution)
        weights[start: start + 3] += 2.0
        weights[end - 3: end] += 2.0
        for peak in peaks:
            weights[peak - 2: peak + 2] += 2.0
        return weights

    def try_fit(self, task: SSUTask) -> typing.Tuple[str, typing.Union[SSUTask, SSUResult]] :
        history = []
        if task.resolver_setting is None:
            setting = SSUAlgorithmSetting()
        else:
            assert isinstance(task.resolver_setting, SSUAlgorithmSetting)
            setting = task.resolver_setting
        classes = np.expand_dims(np.expand_dims(task.sample.classes_φ, 0), 0).repeat(task.n_components, 1)
        distribution_class = DISTRIBUTION_CLASS_MAP[task.distribution_type]
        distance_func = get_distance_function(setting.distance)

        start_time = time.time()
        self.on_fitting_started()

        def closure(x):
            x = x.reshape((1, distribution_class.N_PARAMS+1, task.n_components))
            proportions, components, (m, v, s, k) = distribution_class.interpret(x, classes, task.sample.interval_φ)
            pred_distribution = (proportions[0] @ components[0]).squeeze()
            distance = distance_func(pred_distribution, task.sample.distribution)
            return distance

        def local_callback(x, *addtional):
            x = x.reshape((1, distribution_class.N_PARAMS+1, task.n_components))
            history.append(x)
            self.local_iteration_callback(x)

        initial_guess = task.initial_guess
        if task.initial_guess is None:
            initial_guess = np.expand_dims(distribution_class.get_defaults(task.n_components), 0)

        GO_options = {"maxiter": setting.GO_minimizer_max_niter, "disp": False}
        LO_options = {"maxiter": setting.LO_max_niter, "disp": False}

        if setting.try_GO:
            GO_minimizer_kwargs = \
                dict(method=setting.minimizer,
                     callback=local_callback,
                     options=GO_options)
            GO_result = \
                basinhopping(closure, x0=initial_guess,
                            minimizer_kwargs=GO_minimizer_kwargs,
                            callback=self.global_iteration_callback,
                            niter_success=setting.GO_success_niter,
                            niter=setting.GO_max_niter,
                            stepsize=setting.GO_step)

            if GO_result.lowest_optimization_result.success or \
                    GO_result.lowest_optimization_result.status == 9:
                self.on_global_fitting_succeeded(GO_result)
                initial_guess = GO_result.x
            else:
                self.on_global_fitting_failed(GO_result)
                self.on_fitting_finished()
                return GO_result.message, task

        LO_result = minimize(
            closure,
            method=setting.minimizer,
            x0=initial_guess,
            callback=local_callback,
            options=LO_options) # type: OptimizeResult
        # Judge if the final fitting succeed
        if LO_result.success or LO_result.status == 9:
            finish_time = time.time()
            self.on_fitting_finished()
            time_spent = finish_time - start_time
            func_args = np.reshape(LO_result.x, (1, distribution_class.N_PARAMS+1, task.n_components))
            fitting_result = SSUResult(task, func_args, history=history, time_spent=time_spent)
            self.on_fitting_succeeded(LO_result, fitting_result)
            return LO_result.message, fitting_result
        else:
            self.on_final_fitting_failed(LO_result)
            self.on_fitting_finished()
            return LO_result.message, task


class BackgroundWorker(QObject):
    task_succeeded = Signal(SSUResult)
    task_failed = Signal(str, SSUTask)
    logger = logging.getLogger("QGrain")

    def __init__(self):
        super().__init__()

    def on_task_started(self, task: SSUTask):
        resolver = BasicResolver()
        try:
            message, result = resolver.try_fit(task)
            if isinstance(result, SSUResult):
                self.task_succeeded.emit(result)
            else:
                self.task_failed.emit(message, task)
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.task_failed.emit("An unknown exception was raised. Please check the logs for more details.", task)

class AsyncWorker(QObject):
    task_started = Signal(SSUTask)
    def __init__(self):
        super().__init__()
        self.background_worker = BackgroundWorker()
        self.working_thread = QThread()
        self.background_worker.moveToThread(self.working_thread)
        self.task_started.connect(self.background_worker.on_task_started)
        self.background_worker.task_failed.connect(self.on_task_failed)
        self.background_worker.task_succeeded.connect(self.on_task_succeeded)
        self.working_thread.start()

    def on_task_succeeded(self, fitting_result: SSUResult):
        pass

    def on_task_failed(self, failed_info, task):
        pass

    @Slot()
    def execute_task(self, task: SSUTask):
        self.task_started.emit(task)


def try_sample(
        sample: GrainSizeSample,
        distribution_type: DistributionType,
        n_components: int,
        resolver_setting: SSUAlgorithmSetting = None,
        initial_guess=None):
    task = SSUTask(sample, distribution_type, n_components, resolver_setting, initial_guess)
    resolver = BasicResolver()
    message, result = resolver.try_fit(task)
    return result


def try_dataset(
        dataset: GrainSizeDataset,
        distribution_type: DistributionType,
        n_components: int,
        resolver_setting: SSUAlgorithmSetting = None,
        initial_guess: np.ndarray = None,
        n_processes: int = 1):
    multiprocessing.freeze_support()
    pool = multiprocessing.Pool(n_processes)
    tasks = [SSUTask(sample, distribution_type, n_components, resolver_setting, initial_guess) for sample in dataset.samples]
    def run_task(task: SSUTask):
        resolver = BasicResolver()
        message, result = resolver.try_fit(task)
        return result
    results = pool.map(run_task, tasks)
    suceeded_results = [] # type: list[SSUResult]
    failed_tasks = [] # type: list[SSUTask]
    for result, task in zip(results, tasks):
        if isinstance(result, SSUResult):
            suceeded_results.append(suceeded_results)
        else:
            failed_tasks.append(task)
    return suceeded_results, failed_tasks



__all__ = ["UDMSettingDialog"]

import numpy as np
from PySide6 import QtCore, QtWidgets

from ..udm import UDMAlgorithmSetting
import torch

class UDMSettingDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent, f=QtCore.Qt.Window)
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.tr("UDM Settings"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.device_label = QtWidgets.QLabel(self.tr("Device"))
        self.device_label.setToolTip(self.tr("The neural netowrk framwork, pytorch, also can use the GPU of NVIDIA to do calculations."))
        self.device_combo_box = QtWidgets.QComboBox()
        self.device_combo_box.addItem("CPU")
        if torch.cuda.is_available():
            self.device_combo_box.addItem("CUDA")
            self.device_combo_box.setCurrentText("CUDA")
        self.main_layout.addWidget(self.device_label, 0, 0)
        self.main_layout.addWidget(self.device_combo_box, 0, 1)

        self.pretrain_epochs_label = QtWidgets.QLabel(self.tr("Pretrain Epochs"))
        self.pretrain_epochs_label.setToolTip(self.tr("The number of epochs before formal training. The pretrain process only update the proportions of end members."))
        self.pretrain_epochs_input = QtWidgets.QSpinBox()
        self.pretrain_epochs_input.setRange(10, 100000)
        self.pretrain_epochs_input.setValue(400)
        self.main_layout.addWidget(self.pretrain_epochs_label, 1, 0)
        self.main_layout.addWidget(self.pretrain_epochs_input, 1, 1)

        self.min_epochs_label = QtWidgets.QLabel(self.tr("Minimum Number of Epochs"))
        self.min_epochs_label.setToolTip(self.tr("Minimum number of epochs to be performed."))
        self.min_epochs_input = QtWidgets.QSpinBox()
        self.min_epochs_input.setRange(10, 10000)
        self.min_epochs_input.setValue(2000)
        self.main_layout.addWidget(self.min_epochs_label, 2, 0)
        self.main_layout.addWidget(self.min_epochs_input, 2, 1)

        self.max_epochs_label = QtWidgets.QLabel(self.tr("Maximum Number of Epochs"))
        self.max_epochs_label.setToolTip(self.tr("Maximum number of epochs to be performed."))
        self.max_epochs_input = QtWidgets.QSpinBox()
        self.max_epochs_input.setRange(100, 100000)
        self.max_epochs_input.setValue(5000)
        self.main_layout.addWidget(self.max_epochs_label, 3, 0)
        self.main_layout.addWidget(self.max_epochs_input, 3, 1)

        self.precision_label = QtWidgets.QLabel(self.tr("Precision"))
        self.precision_label.setToolTip(self.tr("It controls the precision for the value of loss function in the stopping criterion."))
        self.precision_input = QtWidgets.QDoubleSpinBox()
        self.precision_input.setRange(2.0, 20.0)
        self.precision_input.setValue(10.0)
        self.main_layout.addWidget(self.precision_label, 4, 0)
        self.main_layout.addWidget(self.precision_input, 4, 1)

        self.learning_rate_label = QtWidgets.QLabel(self.tr("Learning Rate (x10<sup>-3</sup>)"))
        self.learning_rate_label.setToolTip(self.tr("The learning rate of the neural network to update its weights from gradient."))
        self.learning_rate_input = QtWidgets.QDoubleSpinBox()
        self.learning_rate_input.setDecimals(3)
        self.learning_rate_input.setRange(0.001, 1000)
        self.learning_rate_input.setValue(5)
        self.main_layout.addWidget(self.learning_rate_label, 5, 0)
        self.main_layout.addWidget(self.learning_rate_input, 5, 1)

        self.beta1_label = QtWidgets.QLabel(self.tr("Beta 1"))
        self.beta1_label.setToolTip(self.tr("Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta1_input = QtWidgets.QDoubleSpinBox()
        self.beta1_input.setDecimals(4)
        self.beta1_input.setRange(0.0001, 0.9999)
        self.beta1_input.setValue(0.8000)
        self.main_layout.addWidget(self.beta1_label, 6, 0)
        self.main_layout.addWidget(self.beta1_input, 6, 1)

        self.beta2_label = QtWidgets.QLabel(self.tr("Beta 2"))
        self.beta2_label.setToolTip(self.tr("Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta2_input = QtWidgets.QDoubleSpinBox()
        self.beta2_input.setDecimals(4)
        self.beta2_input.setRange(0.0001, 0.9999)
        self.beta2_input.setValue(0.5000)
        self.main_layout.addWidget(self.beta2_label, 7, 0)
        self.main_layout.addWidget(self.beta2_input, 7, 1)

        self.constraint_level_label = QtWidgets.QLabel(self.tr("Constraint Level"))
        self.constraint_level_label.setToolTip(self.tr("It controls the constraint intensity of the end member diversity between different samples. When the constraint level is high, the end members of different samples tend to be equal."))
        self.constraint_level_input = QtWidgets.QDoubleSpinBox()
        self.constraint_level_input.setDecimals(4)
        self.constraint_level_input.setRange(-10, 10.0)
        self.constraint_level_input.setValue(2.0)
        self.main_layout.addWidget(self.constraint_level_label, 8, 0)
        self.main_layout.addWidget(self.constraint_level_input, 8, 1)

    @property
    def setting(self):
        devices = ["cpu", "cuda"]
        setting = UDMAlgorithmSetting(
            device=devices[self.device_combo_box.currentIndex()],
            pretrain_epochs=self.pretrain_epochs_input.value(),
            min_epochs=self.min_epochs_input.value(),
            max_epochs=self.max_epochs_input.value(),
            precision=self.precision_input.value(),
            learning_rate=self.learning_rate_input.value() / 1000.0,
            betas=(self.beta1_input.value(), self.beta2_input.value()),
            constraint_level=self.constraint_level_input.value())
        return setting

    @setting.setter
    def setting(self, setting: UDMAlgorithmSetting):
        device_map = {"cpu": 0, "cuda": 1}
        self.device_combo_box.setCurrentIndex(device_map[setting.device])
        self.pretrain_epochs_input.setValue(setting.pretrain_epochs)
        self.min_epochs_input.setValue(setting.min_epochs)
        self.max_epochs_input.setValue(setting.max_epochs)
        self.precision_input.setValue(setting.precision)
        self.learning_rate_input.setValue(setting.learning_rate*1000.0)
        beta1, beta2 = setting.betas
        self.beta1_input.setValue(beta1)
        self.beta2_input.setValue(beta2)
        self.constraint_level_input.setValue(setting.constraint_level)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("UDM Settings"))
        self.device_label.setText(self.tr("Device"))
        self.device_label.setToolTip(self.tr("The neural netowrk framwork, pytorch, also can use the GPU of NVIDIA to do calculations."))
        self.pretrain_epochs_label.setText(self.tr("Pretrain Epochs"))
        self.pretrain_epochs_label.setToolTip(self.tr("The number of epochs before formal training. The pretrain process only update the proportions of end members."))
        self.min_epochs_label.setText(self.tr("Minimum Number of Epochs"))
        self.min_epochs_label.setToolTip(self.tr("Minimum number of epochs to be performed."))
        self.max_epochs_label.setText(self.tr("Maximum Number of Epochs"))
        self.max_epochs_label.setToolTip(self.tr("Maximum number of epochs to be performed."))
        self.precision_label.setText(self.tr("Precision"))
        self.precision_label.setToolTip(self.tr("It controls the precision for the value of loss function in the stopping criterion."))
        self.learning_rate_label.setText(self.tr("Learning Rate (x10<sup>-3</sup>)"))
        self.learning_rate_label.setToolTip(self.tr("The learning rate of the neural network to update its weights from gradient."))
        self.beta1_label.setText(self.tr("Beta 1"))
        self.beta1_label.setToolTip(self.tr("Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta2_label.setText(self.tr("Beta 2"))
        self.beta2_label.setToolTip(self.tr("Betas are the coefficients used for computing running averages of gradient and its square."))
        self.constraint_level_label.setText(self.tr("Constraint Level"))
        self.constraint_level_label.setToolTip("It controls the constraint intensity of the end member diversity between different samples. When the constraint level is high, the end members of different samples tend to be equal.")



import typing

import numpy as np
from PySide6 import QtCore, QtWidgets

from ..chart.DistributionChart import DistributionChart
from ..model import GrainSizeDataset
from ..ssu import AsyncWorker, DistributionType, SSUResult, SSUTask
from .ParameterEditor import ParameterEditor
from .SSUReferenceViewer import SSUReferenceViewer
from .SSUResultViewer import SSUResultViewer
from .SSUSettingDialog import SSUSettingDialog


class SSUAnalyzer(QtWidgets.QWidget):
    SUPPORT_DISTRIBUTIONS = (
        DistributionType.Normal,
        DistributionType.SkewNormal,
        DistributionType.Weibull,
        DistributionType.GeneralWeibull)
    def __init__(self, setting_dialog: SSUSettingDialog, parameter_editor: ParameterEditor, parent=None):
        super().__init__(parent=parent)
        assert setting_dialog is not None
        assert parameter_editor is not None
        self.setting_dialog = setting_dialog
        self.parameter_editor = parameter_editor
        self.async_worker = AsyncWorker()
        self.async_worker.background_worker.task_succeeded.connect(self.on_fitting_succeeded)
        self.async_worker.background_worker.task_failed.connect(self.on_fitting_failed)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.__dataset = None
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.tr("SSU Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        # control group
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.distribution_label = QtWidgets.QLabel(self.tr("Distribution Type"))
        self.distribution_combo_box = QtWidgets.QComboBox()
        self.distribution_combo_box.addItems([distribution_type.value for distribution_type in self.SUPPORT_DISTRIBUTIONS])
        self.component_number_label = QtWidgets.QLabel(self.tr("Number of Components"))
        self.n_components_input = QtWidgets.QSpinBox()
        self.n_components_input.setRange(1, 10)
        self.n_components_input.setValue(3)
        self.control_layout.addWidget(self.distribution_label, 0, 0)
        self.control_layout.addWidget(self.distribution_combo_box, 0, 1)
        self.control_layout.addWidget(self.component_number_label, 1, 0)
        self.control_layout.addWidget(self.n_components_input, 1, 1)

        self.sample_index_label = QtWidgets.QLabel(self.tr("Sample Index"))
        self.sample_index_input = QtWidgets.QSpinBox()
        self.sample_index_input.valueChanged.connect(self.on_sample_index_changed)
        self.sample_index_input.setEnabled(False)
        self.control_layout.addWidget(self.sample_index_label, 2, 0)
        self.control_layout.addWidget(self.sample_index_input, 2, 1)
        self.sample_name_label = QtWidgets.QLabel(self.tr("Sample Name"))
        self.sample_name_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.control_layout.addWidget(self.sample_name_label, 3, 0)
        self.control_layout.addWidget(self.sample_name_display, 3, 1)

        self.try_fit_button = QtWidgets.QPushButton(self.tr("Try Fit"))
        self.try_fit_button.setEnabled(False)
        self.try_fit_button.clicked.connect(self.on_try_fit_clicked)
        self.edit_parameter_button = QtWidgets.QPushButton(self.tr("Edit Parameters"))
        self.edit_parameter_button.clicked.connect(self.on_edit_parameter_clicked)
        self.control_layout.addWidget(self.try_fit_button, 4, 0)
        self.control_layout.addWidget(self.edit_parameter_button, 4, 1)

        self.try_previous_button = QtWidgets.QPushButton(self.tr("Try Previous"))
        self.try_previous_button.setEnabled(False)
        self.try_previous_button.clicked.connect(self.on_try_previous_clicked)
        self.try_next_button = QtWidgets.QPushButton(self.tr("Try Next"))
        self.try_next_button.setEnabled(False)
        self.try_next_button.clicked.connect(self.on_try_next_clicked)
        self.control_layout.addWidget(self.try_previous_button, 5, 0)
        self.control_layout.addWidget(self.try_next_button, 5, 1)

        # chart group
        self.chart_group = QtWidgets.QGroupBox(self.tr("Chart"))
        self.chart_layout = QtWidgets.QGridLayout(self.chart_group)
        self.result_chart = DistributionChart()
        self.chart_layout.addWidget(self.result_chart, 0, 0)

        # table group
        self.reference_view = SSUReferenceViewer()
        self.reference_view.result_displayed.connect(self.on_result_displayed)
        self.result_view = SSUResultViewer(self.reference_view)
        self.result_view.result_marked.connect(lambda result: self.reference_view.add_references([result]))
        self.result_view.result_displayed.connect(self.on_result_displayed)
        self.result_view.result_referred.connect(self.parameter_editor.refer_ssu_result)
        self.table_tab = QtWidgets.QTabWidget()
        self.table_tab.addTab(self.result_view, self.tr("Result"))
        self.table_tab.addTab(self.reference_view, self.tr("Reference"))

        # pack all group
        self.splitter_1 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.chart_group)
        self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.table_tab)
        self.splitter_2.setStretchFactor(0, 1)
        self.splitter_2.setStretchFactor(1, 2)
        self.main_layout.addWidget(self.splitter_2, 0, 0)

    @property
    def distribution_type(self) -> DistributionType:
        distribution_type = self.SUPPORT_DISTRIBUTIONS[self.distribution_combo_box.currentIndex()]
        return distribution_type

    @property
    def n_components(self) -> int:
        return self.n_components_input.value()

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: GrainSizeDataset):
        self.__dataset = dataset
        self.sample_index_input.setRange(1, dataset.n_samples)
        self.sample_index_input.setValue(1)
        self.on_sample_index_changed(1)
        self.sample_index_input.setEnabled(True)
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)
        self.try_previous_button.setEnabled(True)
        self.try_next_button.setEnabled(True)

    def on_sample_index_changed(self, index):
        self.sample_name_display.setText(self.__dataset.samples[index-1].name)

    def get_task(self, sample_index: int) -> SSUTask:
        sample = self.__dataset.samples[sample_index]
        setting = self.setting_dialog.setting
        if self.parameter_editor.parameter_enabled:
            task = SSUTask(
                sample,
                self.parameter_editor.distribution_type,
                self.parameter_editor.n_components,
                resolver_setting=setting,
                initial_guess=self.parameter_editor.parameters
            )
        else:
            query = self.reference_view.query_reference(sample) # type: SSUResult
            if query is not None:
                task = SSUTask(
                    sample,
                    query.distribution_type,
                    query.n_components,
                    resolver_setting=setting,
                    initial_guess=query.func_args)
            else:
                task = SSUTask(
                    sample,
                    self.distribution_type,
                    self.n_components,
                    resolver_setting=setting)
        return task

    def get_all_tasks(self) -> typing.List[SSUTask]:
        tasks = []
        for i in range(self.__dataset.n_samples):
            task = self.get_task(i)
            tasks.append(task)
        return tasks

    def on_result_displayed(self, result: SSUResult):
        self.result_chart.show_result(result)

    def on_fitting_succeeded(self, fitting_result: SSUResult):
        # update chart
        self.result_chart.show_result(fitting_result)
        self.result_view.add_result(fitting_result)
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)
        self.try_previous_button.setEnabled(True)
        self.try_next_button.setEnabled(True)

    def on_fitting_failed(self, failed_info: str, task: SSUTask):
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)
        self.try_previous_button.setEnabled(True)
        self.try_next_button.setEnabled(True)
        self.show_error(failed_info)

    def do_test(self):
        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button.setEnabled(False)
        self.try_previous_button.setEnabled(False)
        self.try_next_button.setEnabled(False)
        task = self.get_task(self.sample_index_input.value()-1)
        self.async_worker.execute_task(task)

    def on_edit_parameter_clicked(self):
        if self.__dataset is not None:
            sample_index = self.sample_index_input.value()-1
            sample = self.__dataset.samples[sample_index]
            self.parameter_editor.setup_target(sample.classes_μm, sample.distribution)
        self.parameter_editor.show()
        self.parameter_editor.update_chart()

    def on_try_fit_clicked(self):
        self.do_test()

    def on_try_previous_clicked(self):
        current = self.sample_index_input.value()
        if current == 1:
            return
        self.sample_index_input.setValue(current-1)
        self.do_test()

    def on_try_next_clicked(self):
        current = self.sample_index_input.value()
        if current == self.__dataset.n_samples:
            return
        self.sample_index_input.setValue(current+1)
        self.do_test()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Analyzer"))
        self.control_group.setTitle(self.tr("Control"))
        self.distribution_label.setText(self.tr("Distribution Type"))
        self.component_number_label.setText(self.tr("Number of Components"))
        self.sample_index_label.setText(self.tr("Sample Index"))
        self.sample_name_label.setText(self.tr("Sample Name"))
        if not self.__dataset.has_sample:
            self.sample_name_display.setText(self.tr("Unknown"))
        self.try_fit_button.setText(self.tr("Try Fit"))
        self.edit_parameter_button.setText(self.tr("Edit Parameters"))
        self.try_previous_button.setText(self.tr("Try Previous"))
        self.try_next_button.setText(self.tr("Try Next"))
        self.chart_group.setTitle(self.tr("Chart"))
        self.table_tab.setTabText(0, self.tr("Result"))
        self.table_tab.setTabText(1, self.tr("Reference"))
        self.result_view.retranslate()
        self.reference_view.retranslate()



import typing
from enum import Enum, unique

import numpy as np
from scipy.optimize import minimize
from scipy.stats import norm, skewnorm, weibull_min

_INFINITESIMAL = 1e-8


@unique
class DistributionType(Enum):
    Normal = "Normal"
    SkewNormal = "Skew Normal"
    Weibull = "Weibull"
    GeneralWeibull = "General Weibull"


def relu(x):
    return np.maximum(x, 1e-8)


def softmax(x, axis=None):
    x = x - x.max(axis=axis, keepdims=True)
    y = np.exp(x)
    return y / y.sum(axis=axis, keepdims=True)


class Normal:
    NAME = "Normal"
    N_PARAMS = 2
    PARAM_NAMES = ("Location", "Scale")
    PARAM_BOUNDS = ((None, None),
                    (_INFINITESIMAL, None))

    @staticmethod
    def interpret(params: np.ndarray, classes: np.ndarray, interval: np.ndarray):
        n_samples, n_components, n_classes = classes.shape
        # params = params.reshape((n_samples, Normal.N_PARAMS+1, n_components))
        assert params.ndim == 3
        assert params.shape == (n_samples, Normal.N_PARAMS+1, n_components)
        locations = np.expand_dims(params[:, 0, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(params[:, 1, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(params[:, 2, :], axis=1), 1)
        components = norm.pdf(classes, loc=locations, scale=scales) * interval
        mvsk = norm.stats(loc=locations[:, :, 0],scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, mvsk

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((Normal.N_PARAMS+1, n_components))
        defaults[0] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[1] = np.random.random((n_components,)) * 0.1 + 2.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults

    @staticmethod
    def get_initial_guess(mean: float, std: float, skewness: float=0.0) -> np.ndarray:
        return np.array([mean, std])


class SkewNormal:
    NAME = "Skew Normal"
    N_PARAMS = 3
    PARAM_NAMES = ("Shape", "Location", "Scale")
    PARAM_BOUNDS = ((None, None),
                    (None, None),
                    (_INFINITESIMAL, None))

    @staticmethod
    def interpret(params: np.ndarray, classes: np.ndarray, interval: np.ndarray):
        n_samples, n_components, n_classes = classes.shape
        # params = params.reshape((n_samples, SkewNormal.N_PARAMS+1, n_components))
        assert params.ndim == 3
        assert params.shape == (n_samples, SkewNormal.N_PARAMS+1, n_components)
        shapes = np.expand_dims(params[:, 0, :], 2).repeat(n_classes, 2)
        locations = np.expand_dims(params[:, 1, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(params[:, 2, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(params[:, 3, :], axis=1), 1)
        components = skewnorm.pdf(classes, shapes, loc=locations, scale=scales) * interval
        mvsk = skewnorm.stats(shapes[:, :, 0], loc=locations[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, mvsk

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((SkewNormal.N_PARAMS+1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1
        defaults[1] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        defaults[3] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults

    @staticmethod
    def get_initial_guess(mean: float, std: float, skewness: float=0.0) -> np.ndarray:
        x0 = [np.random.rand()*0.1, mean, std]
        def closure(args):
            m, v, s, k = skewnorm.stats(*args, moments="mvsk")
            errors = (mean-m)**2 + (std-np.sqrt(v))**2 + (skewness-k)**2
            return errors
        res = minimize(closure, x0=x0)
        return res.x


class Weibull:
    NAME = "Weibull"
    N_PARAMS = 2
    PARAM_NAMES = ("Shape", "Scale")
    PARAM_BOUNDS = ((_INFINITESIMAL, None),
                    (_INFINITESIMAL, None))

    @staticmethod
    def interpret(params: np.ndarray, classes: np.ndarray, interval: np.ndarray):
        n_samples, n_components, n_classes = classes.shape
        # params = params.reshape((n_samples, Weibull.N_PARAMS+1, n_components))
        assert params.ndim == 3
        assert params.shape == (n_samples, Weibull.N_PARAMS+1, n_components)
        shapes = np.expand_dims(relu(params[:, 0, :]), 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(params[:, 1, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(params[:, 2, :], axis=1), 1)
        components = weibull_min.pdf(classes, shapes, scale=scales) * interval
        mvsk = weibull_min.stats(shapes[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, mvsk

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((Weibull.N_PARAMS+1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1 + 3.60234942
        defaults[1] = np.random.random((n_components,)) * 0.1 + 3.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults

    @staticmethod
    def get_initial_guess(mean: float, std: float, skewness: float=0.0) -> np.ndarray:
        x0 = [np.random.rand()*0.1+3.0, np.random.rand()*0.1+3.0]
        def closure(args):
            m, v, s, k = weibull_min.stats(args[0], scale=args[1], moments="mvsk")
            errors = (mean-m)**2 + (std-np.sqrt(v))**2 + (skewness-k)**2
            return errors
        res = minimize(closure, x0=x0)
        return res.x


class GeneralWeibull:
    NAME = "General Weibull"
    N_PARAMS = 3
    PARAM_NAMES = ("Shape", "Location", "Scale")
    PARAM_BOUNDS = ((_INFINITESIMAL, None),
                    (None, None),
                    (_INFINITESIMAL, None))

    @staticmethod
    def interpret(params: np.ndarray, classes: np.ndarray, interval: np.ndarray):
        n_samples, n_components, n_classes = classes.shape
        # params = params.reshape((n_samples, GeneralWeibull.N_PARAMS+1, n_components))
        assert params.ndim == 3
        assert params.shape == (n_samples, GeneralWeibull.N_PARAMS+1, n_components)
        shapes = np.expand_dims(relu(params[:, 0, :]), 2).repeat(n_classes, 2)
        locations = np.expand_dims(params[:, 1, :], 2).repeat(n_classes, 2)
        scales = np.expand_dims(relu(params[:, 2, :]), 2).repeat(n_classes, 2)
        proportions = np.expand_dims(softmax(params[:, 3, :], axis=1), 1)
        components = weibull_min.pdf(classes, shapes, loc=locations, scale=scales) * interval
        mvsk = weibull_min.stats(shapes[:, :, 0], loc=locations[:, :, 0], scale=scales[:, :, 0], moments="mvsk")
        return proportions, components, mvsk

    @staticmethod
    def get_defaults(n_components: int):
        defaults = np.zeros((GeneralWeibull.N_PARAMS+1, n_components))
        defaults[0] = np.random.random((n_components,)) * 0.1 + 3.60234942
        defaults[1] = np.random.random((n_components,)) * 2.0 + 5.0
        defaults[2] = np.random.random((n_components,)) * 0.1 + 3.0
        defaults[3] = np.random.random((n_components,)) * 0.1 + 2.0
        return defaults

    @staticmethod
    def get_initial_guess(mean: float, std: float, skewness: float=0.0) -> np.ndarray:
        x0 = [np.random.rand()*0.1+3.0, np.random.rand()*0.1+mean-2.0, np.random.rand()*0.1+3.0]
        def closure(args):
            m, v, s, k = weibull_min.stats(*args, moments="mvsk")
            errors = (mean-m)**2 + (std-np.sqrt(v))**2 + (skewness-k)**2
            return errors
        res = minimize(closure, x0=x0)
        return res.x


DISTRIBUTION_CLASS_MAP = {
    DistributionType.Normal: Normal,
    DistributionType.SkewNormal: SkewNormal,
    DistributionType.Weibull: Weibull,
    DistributionType.GeneralWeibull: GeneralWeibull}



import time
import typing

import numpy as np
import torch
from scipy.interpolate import interp1d

from ..model import GrainSizeDataset, GrainSizeSample
from ..ssu import (DISTRIBUTION_CLASS_MAP, DistributionType,
                   SSUAlgorithmSetting, try_sample)
from ..statistic import convert_φ_to_μm, logarithmic
from ._distance import get_distance_func_by_name
from ._kernel import KERNEL_CLASS_MAP, KernelType, Proportion
from ._result import EMMAResult
from ._setting import EMMAAlgorithmSetting


class EMMAModule(torch.nn.Module):
    def __init__(self,
                 n_samples: int,
                 n_members: int,
                 classes_φ: np.ndarray,
                 kernel_type: KernelType,
                 params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = len(classes_φ)
        self.interval = np.abs((classes_φ[0]-classes_φ[-1]) / (classes_φ.shape[0]-1))
        self.classes = torch.nn.Parameter(torch.from_numpy(classes_φ).repeat(1, n_members, 1), requires_grad=False)
        self.kernel_type = kernel_type
        kernel_class = KERNEL_CLASS_MAP[kernel_type]
        self.proportions = Proportion(n_samples, n_members)
        self.end_members = kernel_class(1, self.n_members, self.n_classes, params)

    def forward(self):
        # n_samples x n_members
        proportions = self.proportions().squeeze(1)
        # n_members x n_classes
        end_members = self.end_members(self.classes, self.interval).squeeze(0)
        # n_samples x n_classes
        # distributions = proportions @ end_members
        return proportions, end_members


class EMMAResolver:
    def __init__(self):
        pass

    def try_fit(self, dataset: GrainSizeDataset,
                kernel_type: KernelType,
                n_members: int,
                resolver_setting: EMMAAlgorithmSetting=None,
                parameters=None,
                update_end_members = True):
        if resolver_setting is None:
            s = EMMAAlgorithmSetting(max_epochs=2000, precision=6, learning_rate=5e-3)
        else:
            assert isinstance(resolver_setting, EMMAAlgorithmSetting)
            s = resolver_setting

        X = torch.from_numpy(dataset.distribution_matrix.astype(np.float32)).to(s.device)
        classes_φ = dataset.classes_φ.astype(np.float32)
        emma = EMMAModule(dataset.n_samples, n_members, classes_φ, kernel_type, parameters).to(s.device)

        distance_func = get_distance_func_by_name(s.distance)
        optimizer = torch.optim.Adam(emma.parameters(), lr=s.learning_rate, betas=s.betas)
        loss_series = []
        history = []
        start = time.time()

        emma.end_members.requires_grad = False
        emma.end_members.params.requires_grad = False
        for pretrain_epoch in range(s.pretrain_epochs):
            proportions, end_members = emma()
            X_hat = proportions @ end_members
            loss = distance_func(X_hat, X)
            loss_series.append(loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            history.append((proportions.detach().cpu().numpy(), end_members.detach().cpu().numpy()))

        emma.end_members.requires_grad = update_end_members
        emma.end_members.params.requires_grad = update_end_members
        epochs = 0
        while True:
            # train
            proportions, end_members = emma()
            X_hat = proportions @ end_members
            loss = distance_func(X_hat, X)
            loss_series.append(loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            history.append((proportions.detach().cpu().numpy(), end_members.detach().cpu().numpy()))
            epochs += 1
            if epochs < s.min_epochs:
                continue
            delta_loss = np.mean(loss_series[-100:-80])-np.mean(loss_series[-20:])
            if delta_loss < 10**(-s.precision):
                break
            if epochs > s.max_epochs:
                break
        if resolver_setting.device == "cuda":
            torch.cuda.synchronize()
        time_spent = time.time() - start
        result = EMMAResult(
            dataset,
            kernel_type,
            n_members,
            parameters,
            s,
            proportions.detach().cpu().numpy(),
            end_members.detach().cpu().numpy(),
            time_spent,
            history)
        return result



import time
import typing

import numpy as np
import torch

from ..emma import KERNEL_CLASS_MAP, KernelType, Proportion
from ..model import GrainSizeDataset
from ._result import UDMResult
from ._setting import UDMAlgorithmSetting


class UDMModule(torch.nn.Module):
    def __init__(self,
                 n_samples: int,
                 n_components: int,
                 classes_φ: np.ndarray,
                 kernel_type: KernelType,
                 parameters: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_components = n_components
        self.n_classes = len(classes_φ)
        self.interval = np.abs((classes_φ[0]-classes_φ[-1]) / (classes_φ.shape[0]-1))
        self.classes = torch.nn.Parameter(torch.from_numpy(classes_φ).repeat(n_samples, n_components, 1), requires_grad=False)
        self.kernel_type = kernel_type
        kernel_class = KERNEL_CLASS_MAP[kernel_type]
        self.proportions = Proportion(n_samples, n_components)
        self.components = kernel_class(n_samples, self.n_components, self.n_classes, parameters)

    def forward(self):
        # n_samples x 1 x n_members
        proportions = self.proportions()
        # n_samples x n_members x n_classes
        components = self.components(self.classes, self.interval)
        return proportions, components


class UDMResolver:
    def __init__(self):
        pass

    def try_fit(self, dataset: GrainSizeDataset,
                kernel_type: KernelType,
                n_components: int,
                resolver_setting: UDMAlgorithmSetting = None,
                parameters: np.ndarray = None):
        if resolver_setting is None:
            s = UDMAlgorithmSetting()
        else:
            assert isinstance(resolver_setting, UDMAlgorithmSetting)
            s = resolver_setting

        X = torch.from_numpy(dataset.distribution_matrix.astype(np.float32)).to(s.device)
        classes_φ = dataset.classes_φ.astype(np.float32)
        udm = UDMModule(dataset.n_samples, n_components, classes_φ, kernel_type, parameters).to(s.device)
        optimizer = torch.optim.Adam(udm.parameters(), lr=s.learning_rate, betas=s.betas)

        start = time.time()
        distribution_loss_series = []
        component_loss_series = []
        history = []
        udm.components.requires_grad = False
        udm.components.params.requires_grad = False
        for pretrain_epoch in range(s.pretrain_epochs):
            proportions, components = udm()
            X_hat = (proportions @ components).squeeze(1)
            distribution_loss = torch.log10(torch.mean(torch.square(X - X_hat)))
            distribution_loss_series.append(distribution_loss.item())
            component_loss_series.append(0.0)
            optimizer.zero_grad()
            distribution_loss.backward()
            optimizer.step()
            params = torch.cat([udm.components.params, udm.proportions.params], dim=1).detach().cpu().numpy()
            history.append(params)

        udm.components.requires_grad = True
        udm.components.params.requires_grad = True
        for epoch in range(s.max_epochs):
            # train
            proportions, components = udm()
            X_hat = (proportions @ components).squeeze(1)
            distribution_loss = torch.log10(torch.mean(torch.square(X_hat - X)))
            component_loss = torch.mean(torch.std(components, dim=0))
            loss = distribution_loss + (10**s.constraint_level) * component_loss
            distribution_loss_series.append(distribution_loss.item())
            component_loss_series.append((10**s.constraint_level) * component_loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            params = torch.cat([udm.components.params, udm.proportions.params], dim=1).detach().cpu().numpy()
            history.append(params)

            if np.isnan(loss.item()):
                break

            if epoch > s.min_epochs:
                delta_loss = np.mean(distribution_loss_series[-100:-80])-np.mean(distribution_loss_series[-20:])
                if delta_loss < 10**(-s.precision):
                    break

        if s.device == "cuda":
            torch.cuda.synchronize()
        time_spent = time.time() - start
        final_params = torch.cat([udm.components.params, udm.proportions.params], dim=1).detach().cpu().numpy()
        result = UDMResult(
            dataset, kernel_type, n_components,
            parameters,
            s,
            np.array(distribution_loss_series),
            np.array(component_loss_series),
            time_spent,
            final_params,
            history)
        return result



__all__ = [
    "normal_color",
    "highlight_color",
    "synchronize_theme",
    "setup_matplotlib"]

import os

import matplotlib.pyplot as plt


def normal_color():
    return os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]

def highlight_color():
    return os.environ["QTMATERIAL_PRIMARYCOLOR"]

def backgroud_color():
    return os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]

def synchronize_theme():
    plt.rcParams["axes.facecolor"] = os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]
    plt.rcParams["figure.facecolor"] = os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]
    plt.rcParams["axes.edgecolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["axes.titlecolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["axes.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["grid.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["legend.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["xtick.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["xtick.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["ytick.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["ytick.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]

    plt.rcParams["savefig.dpi"] = 300.0
    plt.rcParams["savefig.transparent"] = True
    plt.rcParams["figure.max_open_warning"] = False

def setup_matplotlib():
    plt.style.use(["science", "no-latex"])
    plt.set_cmap("tab10")
    plt.rcParams['axes.unicode_minus'] = False
    synchronize_theme()



__all__ = ["SSUSettingDialog"]

import numpy as np
from PySide6 import QtCore, QtWidgets

from ..ssu import SSUAlgorithmSetting, built_in_distances, built_in_minimizers


class SSUSettingDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent, f=QtCore.Qt.Window)
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.tr("SSU Settings"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        # distance
        self.distance_label = QtWidgets.QLabel(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The function to calculate the difference (on the contrary, similarity) between two samples."))
        self.distance_combo_box = QtWidgets.QComboBox()
        self.distance_combo_box.addItems(built_in_distances)
        self.distance_combo_box.setCurrentText("log10MSE")
        self.main_layout.addWidget(self.distance_label, 0, 0)
        self.main_layout.addWidget(self.distance_combo_box, 0, 1)
        # minimizer
        self.minimizer_label = QtWidgets.QLabel(self.tr("Minimizer"))
        self.minimizer_label.setToolTip(self.tr("The minimizer to find the minimum value of the distance function."))
        self.minimizer_combo_box = QtWidgets.QComboBox()
        self.minimizer_combo_box.addItems(built_in_minimizers)
        self.minimizer_combo_box.setCurrentText("SLSQP")
        self.main_layout.addWidget(self.minimizer_label, 1, 0)
        self.main_layout.addWidget(self.minimizer_combo_box, 1, 1)
        # try_GO
        self.try_GO_checkbox = QtWidgets.QCheckBox(self.tr("Global Optimization"))
        self.try_GO_checkbox.setChecked(False)
        self.try_GO_checkbox.setToolTip(self.tr("Whether to try global optimization (GO) before local optimization (LO)."))
        self.main_layout.addWidget(self.try_GO_checkbox, 2, 0, 1, 2)
        # GO_max_niter
        self.GO_max_niter_label = QtWidgets.QLabel(self.tr("[GO] Maximum Number of Iterations"))
        self.GO_max_niter_label.setToolTip(self.tr("Maximum number of iterations of global optimization for termination."))
        self.GO_max_niter_input = QtWidgets.QSpinBox()
        self.GO_max_niter_input.setRange(0, 10000)
        self.GO_max_niter_input.setValue(100)
        self.main_layout.addWidget(self.GO_max_niter_label, 3, 0)
        self.main_layout.addWidget(self.GO_max_niter_input, 3, 1)
        # GO_success_niter
        self.GO_success_niter_label = QtWidgets.QLabel(self.tr("[GO] Success Number of Iterations"))
        self.GO_success_niter_label.setToolTip(self.tr("The number of iteration that reaching the same local optimal value for termination."))
        self.GO_success_niter_input = QtWidgets.QSpinBox()
        self.GO_success_niter_input.setRange(1, 100)
        self.GO_success_niter_input.setValue(5)
        self.main_layout.addWidget(self.GO_success_niter_label, 4, 0)
        self.main_layout.addWidget(self.GO_success_niter_input, 4, 1)
        # GO_step
        self.GO_step_label = QtWidgets.QLabel(self.tr("[GO] Step Size"))
        self.GO_step_label.setToolTip(self.tr("The step size of searching global optimal value."))
        self.GO_step_input = QtWidgets.QDoubleSpinBox()
        self.GO_step_input.setRange(0.01, 10)
        self.GO_step_input.setValue(0.1)
        self.main_layout.addWidget(self.GO_step_label, 5, 0)
        self.main_layout.addWidget(self.GO_step_input, 5, 1)
        # GO_minimizer_max_niter
        self.GO_minimizer_max_niter_label = QtWidgets.QLabel(self.tr("[GO] Minimizer Maximum Number of Iterations"))
        self.GO_minimizer_max_niter_label.setToolTip(self.tr("Maximum number of iterations of the minimizer of global optimization."))
        self.GO_minimizer_max_niter_input = QtWidgets.QSpinBox()
        self.GO_minimizer_max_niter_input.setRange(0, 100000)
        self.GO_minimizer_max_niter_input.setValue(500)
        self.main_layout.addWidget(self.GO_minimizer_max_niter_label, 6, 0)
        self.main_layout.addWidget(self.GO_minimizer_max_niter_input, 6, 1)
        # LO_max_niter
        self.LO_max_niter_label = QtWidgets.QLabel(self.tr("[LO] Maximum Number of Iterations"))
        self.LO_max_niter_label.setToolTip(self.tr("Maximum number of iterations of the minimizer of final local optimization."))
        self.LO_max_niter_input = QtWidgets.QSpinBox()
        self.LO_max_niter_input.setRange(100, 100000)
        self.LO_max_niter_input.setValue(1000)
        self.main_layout.addWidget(self.LO_max_niter_label, 7, 0)
        self.main_layout.addWidget(self.LO_max_niter_input, 7, 1)

    @property
    def setting(self) -> SSUAlgorithmSetting:
        setting = SSUAlgorithmSetting(
            distance=self.distance_combo_box.currentText(),
            minimizer=self.minimizer_combo_box.currentText(),
            try_GO=self.try_GO_checkbox.isChecked(),
            GO_max_niter=self.GO_max_niter_input.value(),
            GO_success_niter=self.GO_success_niter_input.value(),
            GO_step=self.GO_step_input.value(),
            GO_minimizer_max_niter=self.GO_minimizer_max_niter_input.value(),
            LO_max_niter=self.LO_max_niter_input.value())
        return setting

    @setting.setter
    def setting(self, setting: SSUAlgorithmSetting):
        self.distance_combo_box.setCurrentText(setting.distance)
        self.minimizer_combo_box.setCurrentText(setting.minimizer)
        self.try_GO_checkbox.setChecked(setting.try_GO)
        self.GO_max_niter_input.setValue(setting.GO_max_niter)
        self.GO_success_niter_input.setValue(setting.GO_success_niter)
        self.GO_step_input.setValue(setting.GO_step)
        self.GO_minimizer_max_niter_input.setValue(setting.GO_minimizer_max_niter)
        self.LO_max_niter_input.setValue(setting.LO_max_niter)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Settings"))
        self.distance_label.setText(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The function to calculate the difference (on the contrary, similarity) between two samples."))
        self.minimizer_label.setText(self.tr("Minimizer"))
        self.minimizer_label.setToolTip(self.tr("The minimizer to find the minimum value of the distance function."))
        self.try_GO_checkbox.setText(self.tr("Global Optimization"))
        self.try_GO_checkbox.setToolTip(self.tr("Whether to try global optimization (GO) before local optimization (LO)."))
        self.GO_max_niter_label.setText(self.tr("[GO] Maximum Number of Iterations"))
        self.GO_max_niter_label.setToolTip(self.tr("Maximum number of iterations of global optimization for termination."))
        self.GO_success_niter_label.setText(self.tr("[GO] Success Number of Iterations"))
        self.GO_success_niter_label.setToolTip(self.tr("The number of iteration that reaching the same local optimal value for termination."))
        self.GO_step_label.setText(self.tr("[GO] Step Size"))
        self.GO_step_label.setToolTip(self.tr("The step size of searching global optimal value."))
        self.GO_minimizer_max_niter_label.setText(self.tr("[GO] Minimizer Maximum Number of Iterations"))
        self.GO_minimizer_max_niter_label.setToolTip(self.tr("Maximum number of iterations of the minimizer of global optimization."))
        self.LO_max_niter_label.setText(self.tr("[LO] Maximum Number of Iterations"))
        self.LO_max_niter_label.setToolTip(self.tr("Maximum number of iterations of the minimizer of final local optimization."))
